#include "suif_inc.h"
#include <map>
#include <algorithm>
#include "dump.h"
#include "build.h"

#ifdef GNUMP
#define value_to_int(v) ((int)mpz_get_si(v))
#else
#define value_to_int(v) ((int)v)
#endif

struct transformer {
    FILE	*out;
    int		 indent;

    CloogOptions *options;
    CloogProgram *program;
    tree_proc	 *tp;
    symtab_ann   *sa;

    map<int, instruction *>		nr2insn;
    map<instruction *, domain_ann *>	insn2dom;
    std::vector<var_sym *>		scat;
    std::vector<var_sym *>		it;

    transformer(tree_proc *t, CloogProgram *p, CloogOptions *o);
    void build(tree_node *root);
    void visit(tree_node *t);

    instruction *scan(clast_term *e);
    instruction *scan(clast_reduction *e);
    instruction *scan(clast_binary *e);
    instruction *scan(clast_expr *e);
    tree_node_list *scan(clast_stmt *s);
    tree_node_list *scan(clast_assignment *s);
    tree_node_list *scan(clast_user_stmt *s);
    tree_node_list *scan(clast_for *s);
    instruction *scan(clast_equation *e);
    tree_node_list *scan(clast_guard *s);

    var_sym* name(const char *name);
};

transformer::transformer(tree_proc *t, CloogProgram *p, CloogOptions *o) : 
    tp(t), program(p)
{
    options = o;
    sa = (symtab_ann *)tp->peek_annote(k_domain_symtab);
}

void transformer::visit(tree_node *t)
{
    if (!t->peek_annote(k_static_root))
	return;
    build(t);
}

static void visit (tree_node *t, void *d)
{
    ((transformer *)d)->visit(t);
}

void build(tree_proc *tp, CloogProgram *program, CloogOptions *options)
{
    transformer b(tp, program, options);
    //unused_start_symtab(tp->scope(), TRUE, FALSE);
    tp->map(::visit, (void *) &b);
    forward_propagate(tp, FPK_ALL);
    fold_all_constants(tp);
    proc_dead_code(tp);
    proc_unused_syms(tp, TRUE, FALSE);
    //unused_end_symtab(tp->scope());
}

void transformer::build(tree_node *root)
{
    block::set_proc(tp);

    program = cloog_program_generate(program,options) ;
  
    out		   = stdout;
    indent	   = 0;
    
    dumper d(tp, NULL);
    d.scan();
    for (int i = 0; i < d.nodes.size(); ++i) {
	domain_ann *dom = d.nodes[i];
	instruction *insn = dom->insn;
	nr2insn[insn->number()] = insn;
	insn2dom[insn] = dom;
    }
    for (int i = 0; i < program->nb_scattdims; ++i)
	scat.push_back(tp->symtab()->new_unique_var(type_signed, "c"));
    for (int i = 0; i < program->names->nb_iterators; ++i)
	it.push_back(tp->symtab()->new_unique_var(type_signed, "i"));

    clast_stmt *stmt = cloog_clast_create(program, options);
    // cloog_program_pprint(stderr, program, options);
    tree_node_list *list = scan(stmt);
    cloog_clast_free(stmt);

    tree_node_list *tnl = ((tree_block*) root)->body();
    tnl->clear();
    tree_node_list_iter tnli(list);
    while (!tnli.is_empty()) {
	tnl->append(tnli.step());
    }
  
    cloog_options_free(options) ;
    cloog_program_free(program) ;
}

instruction *transformer::scan(clast_term *t)
{
    if (t->var)
	return block(block(value_to_int(t->val)) * block(name(t->var))).
		    make_instruction();
    else
	return block(value_to_int(t->val)).make_instruction();
}

var_sym* transformer::name(const char *name)
{
    int number;
    if (sscanf(name, "c%d", &number))
	return scat[number];
    else if (sscanf(name, "i%d", &number))
	return it[number];
    else {
	map<var_sym *, Param_Iterator *>::iterator gi;
	for (gi = sa->vars.begin(); gi != sa->vars.end(); ++gi) {
	    if (!strcmp(gi->second->base_name(), name))
		break;
	}
	assert(gi != sa->vars.end());
	return gi->first;
    }
}

tree_node_list *transformer::scan(clast_assignment *a)
{
    block& b = block(name(a->LHS)).doasignop(aop_eq, block(scan(a->RHS)));
    return b.make_tree_node_list() ;
}

instruction *transformer::scan(clast_reduction *r)
{
    int n = r->n;
    if (r->n == 1)
	return scan(r->elts[0]);

    binary_op bop;
    switch (r->type) {
	case clast_red_sum:
	    bop = bop_add;
	    break;
	case clast_red_max:
	    bop = bop_max;
	    break;
	case clast_red_min:
	    bop = bop_min;
	    break;
	default:
	    assert(0);
    }

    block one(scan(r->elts[n-1]));
    while (--n)
	one.set(block::op(scan(r->elts[n-1]), bop, one));
    return one.make_instruction();
}

instruction *transformer::scan(clast_equation *e)
{
    instruction *l = scan(e->LHS);
    instruction *r = scan(e->RHS);
    binary_op bop = e->sign ? e->sign > 0 ? bop_geq : bop_leq : bop_eq;
    return block::op(l, bop, r).make_instruction();
}

tree_node_list* transformer::scan(clast_guard *g)
{
    int n = g->n;
    block guard(scan(&g->eq[n-1]));
    while (--n)
	guard.set(block::op(scan(&g->eq[n-1]), bop_and, guard));
    block code = block::IF(guard, block(scan(g->then)));
    return code.make_tree_node_list() ;
}

tree_node_list* transformer::scan(clast_for *f)
{
    if (!f->UB) {
	/* If no upper bound, then assume infinite loop */
	/* We create a while loop with a test that simply jumps back to the top */
	tree_node_list* list = new tree_node_list();
	label_sym *top = block::get_proc()->proc_syms()->new_unique_label("L_top");
	label_sym *cont = block::get_proc()->proc_syms()->new_unique_label("L_cont");
	label_sym *brk = block::get_proc()->proc_syms()->new_unique_label("L_brk");
	block body = scan(f->body);
	tree_instr *ti = new tree_instr(new in_bj(io_jmp, top));
	tree_node_list *tnl = new tree_node_list;
	tnl->append(ti);
	tree_loop *tl = new tree_loop(body.make_tree_node_list(), tnl,
				      cont, brk, top);
	list->append(tl);
	return list;
    } else {
	block code = block::FOR(name(f->iterator), 
			  scan(f->LB), scan(f->UB), scan(f->body));
	return code.make_tree_node_list() ;
    }
}

tree_node_list* transformer::scan(clast_user_stmt *u)
{
    instruction *insn = nr2insn[u->statement->number];
    domain_ann *dom = insn2dom[insn];
    tree_node_list *s = scan(u->substitutions);
    block subs(s);
    if (dom) {
	for (int i = 0; i < dom->level; ++i) {
	    if (tp->symtab()->lookup_var(dom->its[i]->name()) == dom->its[i])
		subs.set(block(subs, block(dom->its[i]) = block(it[i])));
	}
    }
    tree_node_list* r = block(subs,block(insn)).make_tree_node_list();
    delete s;
    return r;
}

tree_node_list* transformer::scan(clast_stmt *s)
{
    tree_node_list* list = new tree_node_list();
    for ( ; s; s = s->next) {
	if (CLAST_STMT_IS_A(s, stmt_ass)) {
	    list->append(scan((clast_assignment *) s));
	} else if (CLAST_STMT_IS_A(s, stmt_user)) {
	    list->append(scan((clast_user_stmt *) s));
	} else if (CLAST_STMT_IS_A(s, stmt_for)) {
	    list->append(scan((clast_for *) s));
	} else if (CLAST_STMT_IS_A(s, stmt_guard)) {
	    list->append(scan((clast_guard *) s));
	} else if (CLAST_STMT_IS_A(s, stmt_root)) {
	    ;
	} else {
	    assert(0);
	}
    }
    return list;
}

instruction *transformer::scan(clast_expr *e)
{
    if (!e)
	return NULL;
    switch (e->type) {
    case clast_expr::expr_term:
	return scan((clast_term*) e);
	break;
    case clast_expr::expr_red:
	return scan((clast_reduction*) e);
	break;
    case clast_expr::expr_bin:
	return scan((clast_binary*) e);
	break;
    default:
	assert(0);
    }
}

instruction *transformer::scan(clast_binary *e)
{
    binary_op bop;
    switch (e->type) {
	case clast_bin_div:
	    bop = bop_div;
	    break;
	case clast_bin_mod:
	    bop = bop_rem;
	    break;
	case clast_bin_cdiv:
	    bop = bop_divceil;
	    break;
	case clast_bin_fdiv:
	    bop = bop_divfloor;
	    break;
	default:
	    assert(0);
    }
    block b(block::op(scan(e->LHS), bop, value_to_int(e->RHS)));
    return b.make_instruction();
}
