#include "suif_inc.h"
#include <algorithm>
#include <sstream>
#include "domain.h"

const char * k_static_root;
const char * k_stmt;
const char * k_if_annote;
const char * k_domain_symtab;
const char * k_iteration_domain;

void symtab_ann::add_var(var_sym *var)
{
    if (vars.find(var) == vars.end())
	vars[var] = new Param_Iterator(var->name());
}

void
domain_constructor::recurse(tree_node * tn)
{
    unsigned n = tn->num_child_lists();
    for (unsigned i = 0; i < n; ++i)
	construct(tn->child_list_num(i));
}

SRelation * domain_constructor::reorder(SRelation * c)
{
    Relation r = copy(*(Relation*)c);
    r = Extend_Set(r, level);
    for (int i = 0; i < level; ++i) {
	EQ_Handle e = r.and_with_EQ();
	e.update_coef(r.get_local(sa->vars[its[i]]), -1);
	e.update_coef(r.set_var(i+1), 1);
	r = Project(r, sa->vars[its[i]]);
    }
    return new SRelation(r, sa);
}

/*
 * Reorder the columns in c such that the first columns correspond
 * to the iterators (in order), optionally followed by the array dimensions
 * (in order) and followed by the parameters (in arbitrary order), 
 * columns for iterators are added if needed
 *
 * params is the number of parameters
 *	if -1, then the parameters are counted
 * output is the number of output variables (array dimensions)
 */

named_lin_ineq * 
domain_constructor::reorder(named_lin_ineq * c, int params, int output)
{
    named_lin_ineq * o = new named_lin_ineq(c);
    if (params == -1) {
	params = 0;
	for (int i = 1; i < o->n(); ++i) {
	    var_sym * v = o->names()[i].var();
	    if (find(its.begin(), its.end(), v) == its.end() &&
	        find(out.begin(), out.end(), v) == out.end())
		++params;
	}
    }
    for (int i = 0; i < level; ++i) {
	int pos = o->names().find(its[i]);
	if (pos == -1)
	    o->add_col(name_table_entry(its[i]), i + 1);
	else if (pos != i + 1)
	    o->swap_col(pos, i + 1);
    }
    for (int i = 0; i < output; ++i) {
	int pos = o->names().find(out[i]);
	if (pos != i + level + 1)
	    o->swap_col(pos, i + level + 1);
    }
    return o;
}

stmt_access *domain_constructor::create_access(dep_instr_annote *di, bool write)
{
	instruction *in = di->my_instr;
	assert(in->opcode() == io_array);
	in_array *ina = (in_array *)in;
	assert(ina->base_op().is_instr());
	var_sym *array;
	switch (ina->base_op().instr()->opcode()) {
	case io_ldc: {
		in_ldc *ldc = (in_ldc *) ina->base_op().instr();
		assert(ldc->value().kind() == im_symbol);
		assert(ldc->value().symbol()->kind() == SYM_VAR);
		array = (var_sym *)ldc->value().symbol();
		break;
	}
	case io_cvt: {
		instruction *in = ina->base_op().instr();
		assert(in->src_op(0).kind() == OPER_SYM);
		assert(in->src_op(0).symbol()->kind() == SYM_VAR);
		array = (var_sym *)in->src_op(0).symbol();
		break;
	}
	default:
		assert(0);
	}

	named_lin_ineq *a = new named_lin_ineq();
	array_info_iter iter(di->ai);
	int i;
	for(i = 0; !iter.is_empty(); ++i) {
		if (i >= out.size())
		    out.push_back(sa->symtab->new_unique_var(type_signed));

		access_vector * av = iter.step();
		if (av->too_messy)
			in->print();
		assert(!av->too_messy);
		named_lin_ineq * t;
		*a &= *(t = named_lin_ineq::mk_named_lin_ineq(*av, out[i],
							      FALSE));
		delete t;
		*a &= *(t = named_lin_ineq::mk_named_lin_ineq(*av, out[i],
							      TRUE));
		delete t;
	}

	return new stmt_access(write, a, array, i);
}

/* Returns stmt_access if it's an access already analyzed by SUIF */
stmt_access *domain_constructor::check_access(instruction *in, bool write)
{
	dep_instr_annote *di =
		(dep_instr_annote *)in->peek_annote(k_dep_instr_annote);
	if (!di)
		return NULL;

	return create_access(di, write);
}

stmt_node *domain_constructor::extract_stmt_node(operand &op, boolean write)
{
	if (op.is_expr())
		return extract_stmt_node(op.instr());

	if (!op.is_symbol()) {
		op.print(stderr);
		assert(0);
	}

	if (op.symbol()->type()->is_ptr())
		return NULL;
	named_lin_ineq *a = new named_lin_ineq();
	var_sym *array = op.symbol();
	int output = 0;
	if (sa->vars.find(array) != sa->vars.end()) {
		/* an iterator or a parameter */
		/* set inequalities to "out[0] = array */
		assert(!write);
		if (out.size() == 0)
			out.push_back(sa->symtab->new_unique_var(type_signed));
		a->ineqs().init(2,1);
		a->names().init(1);
		a->add_col(name_table_entry(array), 1);
		a->add_col(name_table_entry(out[0]), 2);
		a->ineqs()[0][1] = 1;
		a->ineqs()[0][2] = -1;
		a->ineqs()[1][1] = -1;
		a->ineqs()[1][2] = 1;
		array = NULL;
		output = 1;
	}
	return new stmt_access(write, a, array, output);
}

static stmt *add_access(stmt *s, stmt_node *out)
{
	stmt_access *acc = dynamic_cast<stmt_access *>(out);
	if (!acc) {
		delete s;
		return NULL;
	}
	acc->write = true;
	s->out.push_back(acc);
	return s;
}

/*
 * Returns NULL if instruction is not ok
 * Checks currently incomplete
 */
stmt *domain_constructor::extract_stmt(tree_instr *t)
{
	stmt_node *out = NULL;
	instruction *i = t->instr();
	stmt_node *node;

	if (i->opcode() == io_str || i->opcode() == io_memcpy) {
		out = extract_stmt_node(i->src_op(0).instr());
		if (!out)
			return NULL;
		operand r = i->src_op(1);
		node = extract_stmt_node(r, false);
	} else
		node = extract_stmt_node(i);

	if (!node) {
		if (out)
			delete out;
		return NULL;
	}

	stmt *s = new stmt;
	s->tree = node;

	if (out)
		s = add_access(s, out);
	if (!s)
		return NULL;

	for (unsigned n = 0; n < i->num_dsts(); n++) {
		operand r = i->dst_op(n);
		if (r.is_null())
			continue;
		stmt_node *dst = extract_stmt_node(r, true);
		s = add_access(s, dst);
		if (!s)
			return NULL;
	}

	return s;
}

static const char *instr_name(instruction *i)
{
	if (i->opcode() == io_cal) {
		in_cal *cal = dynamic_cast<in_cal *>(i);
		in_ldc *ldc = dynamic_cast<in_ldc *>(cal->addr_op().instr());
		assert(ldc->value().symbol()->is_proc());
		proc_sym *proc = dynamic_cast<proc_sym *>(ldc->value().symbol());
		return proc->name();
	}
	if (i->opcode() == io_str)
		return "";
	if (i->opcode() == io_ldc) {
		in_ldc *ins = dynamic_cast<in_ldc *>(i);
		std::ostringstream strm;
		strm << ins->value().integer();
		return strdup(strm.str().c_str());
	}
	if (i->opcode() == io_add)
		return "+";
	if (i->opcode() == io_sub)
		return "-";
	if (i->opcode() == io_mul)
		return "*";
	if (i->opcode() == io_div)
		return "/";
	fprintf(stderr, "unrecognized instruction\n");
	i->print(stderr);
	return NULL;
}

stmt_node *domain_constructor::extract_stmt_node(instruction *i)
{
	stmt_node *node;

	node = check_access(i, false);
	if (node)
		return node;

	if (i->opcode() == io_lod) {
		operand r = i->src_op(0);
		return extract_stmt_node(r, false);
	}

	if (i->opcode() == io_cal) {
		in_cal *cal = dynamic_cast<in_cal *>(i);
		return extract_stmt_node_cal(cal);
	}

	const char *name = instr_name(i);
	if (!name)
		return NULL;

	stmt_call *call = new stmt_call(name);

	for (unsigned n = 0; n < i->num_srcs(); n++) {
		operand r = i->src_op(n);
		node = extract_stmt_node(r, false);
		if (!node) {
			delete call;
			return NULL;
		}
		call->arg.push_back(node);
	}

	return call;
}

stmt_node *domain_constructor::extract_stmt_node_cal(in_cal *cal)
{
	const char *name = instr_name(cal);
	if (!name)
		return NULL;

	stmt_call *call = new stmt_call(name);

	in_ldc *ldc = dynamic_cast<in_ldc *>(cal->addr_op().instr());
	assert(ldc->value().symbol()->is_proc());
	proc_sym *proc = dynamic_cast<proc_sym *>(ldc->value().symbol());
	func_type *type = proc->type();
	if (!type->args_known()) {
		type = type->clone();
		type->set_num_args(cal->num_args());
		proc->set_type(type);
		for (unsigned n = 0; n < type->num_args(); ++n)
			type->set_arg_type(n, type_signed);
		type->set_return_type(type_signed);
		tp->proc()->file()->parent()->globals()->add_type(type);
	}
	for (unsigned n = 0; n < type->num_args(); ++n) {
		stmt_node *node;
		type_node *atype = type->args_known() ? type->arg_type(n) : NULL;
		operand op = cal->argument(n);
		if (!atype || !atype->is_ptr()) {
			node = extract_stmt_node(op, false);
			if (!node) {
				delete call;
				return NULL;
			}
			call->arg.push_back(node);
			continue;
		}
		/* if the function expects a pointer
		 * consider data of which the address is taken
		 * to be input if the type is pointer to const
		 * and output otherwise.
		 */
		ptr_type *ptr = static_cast<ptr_type *>(atype);
		type_node *base = ptr->ref_type();
		boolean write = !base->is_const();

		if (op.instr()->opcode() == io_cvt)
			op = op.instr()->src_op(0);
		node = check_access(op.instr(), write);
		if (node) {
			call->arg.push_back(node);
			continue;
		}

		in_ldc *ldc = dynamic_cast<in_ldc *>(op.instr());
		assert(ldc);
		assert(ldc->value().kind() == im_symbol);
		assert(ldc->value().symbol()->kind() == SYM_VAR);
		var_sym *var = static_cast<var_sym*>(ldc->value().symbol());

		named_lin_ineq *a = new named_lin_ineq();
		node = new stmt_access(write, a, var, 0);
		call->arg.push_back(node);
	}

	return call;
}

void domain_constructor::set_prefix_and_iterator(int prefix_value, var_sym *iterator)
{
    if (plevel >= prefix.size())
	prefix.push_back(prefix_value);
    else
	prefix[plevel] = prefix_value;

    if (!iterator && level < its.size() && its[level]->parent() == sa->symtab)
	return;
    if (!iterator)
	iterator = sa->symtab->new_unique_var(type_signed);
    if (level >= its.size())
	its.push_back(iterator);
    else
	its[level] = iterator;
    sa->add_var(its[level]);
    sa->vars[its[level]]->iterator = true;
}

void
domain_constructor::construct(tree_node * tn)
{
    switch (tn->kind()) {
    case TREE_INSTR: {
	tree_instr *in = (tree_instr*)tn;
	switch (in->instr()->opcode()) {
	case io_mrk:
	    break;
	default:
	    assert(0);
	case io_memcpy:
	case io_cpy:
	case io_str:
	case io_lod:
	case io_cvt:
	case io_ldc:
	case io_add:
	case io_sub:
	case io_div:
	case io_cal:
	    root = in->instr();
	    stmt *s = NULL;

	    annote_list_iter i(root->annotes());
	    while (!i.is_empty()) {
		annote *an = i.step();
		if (an->name() == k_stmt) {
			s = (stmt *)an->data();
			s->reorder(this);
		}
	    }

	    SRelation * o = reorder(c);

	    domain_ann *da;
	    da = new domain_ann(in->instr(), o, level, plevel, its, prefix, s);
	    in->instr()->append_annote(k_iteration_domain, da);
	}
	break;
    }
    case TREE_FOR: {
	SRelation * oldc = new SRelation(*c);
	dep_for_annote * df = 
		(dep_for_annote *)tn->peek_annote(k_dep_for_annote);
	tree_for * tf = df->my_for;
	immed idx(tf->index());
	c->intersect(named_lin_ineq::mk_named_lin_ineq(*df->lb, idx, TRUE));
	c->intersect(named_lin_ineq::mk_named_lin_ineq(*df->ub, idx, FALSE));

	set_prefix_and_iterator(-1, tf->index());

	++level;
	++plevel;
	construct(tf->body());
	--plevel;
	--level;

	delete c;
	c = oldc;
	break;
    }
    case TREE_LOOP: {
	/* We only get here if the loop was determined to be a simple infinite loop */
	SRelation * oldc = new SRelation(*c);
	tree_loop *tl = static_cast<tree_loop *>(tn);

	set_prefix_and_iterator(-1);
	c->set_infinite(its[level]);

	++level;
	++plevel;
	construct(tl->body());
	--plevel;
	--level;

	delete c;
	c = oldc;
	break;
    }
    case TREE_IF: {
	tree_if * ti = (tree_if *)tn;
	SRelation * oldc = c;

	SRelation * s = (SRelation *)tn->peek_annote(k_if_annote);
	c = s->Intersection(oldc);

	construct(ti->then_part());

	delete c;

	if (ti->else_part()->is_empty()) {
	    c = oldc;
	    break;
	}

	c = oldc->Difference(s);

	construct(ti->else_part());

	delete c;

	c = oldc;
	break;
    }
    default:
	recurse(tn);
	break;
    }
}

void
domain_constructor::construct(tree_node_list * tnl)
{
    set_prefix_and_iterator(0);

    tree_node_list_iter tnli(tnl);
    for (int i = 0; !tnli.is_empty(); ++i) {
	tree_node *t = tnli.step();

	SRelation * oldc = new SRelation(*c);

	if (oldstyle)
	    c->fix(its[level], i);
	prefix[plevel] = i;

	if (oldstyle)
	    ++level;
	++plevel;
	construct(t);
	--plevel;
	if (oldstyle)
	    --level;

	delete c;
	c = oldc;
    }
}

void domain_constructor::visit(tree_node *t)
{
    if (!t->peek_annote(k_static_root))
	return;

    c = new SRelation(sa);
    level = 0;
    plevel = 0;
    construct(t);
    delete c;
}

static void visit (tree_node *t, void *d)
{
    ((domain_constructor *)d)->visit(t);
}

void domain_constructor::construct()
{
    tp->map(::visit, (void *) this);
}

boolean guard_constructor::fill_q(Constraint_Handle& h, operand op, int mult)
{
    switch(op.kind()) {
    case OPER_INSTR:
	switch (op.instr()->opcode()) {
	    case io_ldc: {
		in_ldc *ins = dynamic_cast<in_ldc *>(op.instr());
		h.update_const(mult * ins->value().integer());
		break;
	    }
	    case io_add: {
		in_rrr * ir = dynamic_cast<in_rrr *>(op.instr());
		if (!fill_q(h, ir->src1_op(), mult))
		    return false;
		if (!fill_q(h, ir->src2_op(), mult))
		    return false;
		break;
	    }
	    case io_sub: {
		in_rrr * ir = dynamic_cast<in_rrr *>(op.instr());
		if (!fill_q(h, ir->src1_op(), mult))
		    return false;
		if (!fill_q(h, ir->src2_op(), -mult))
		    return false;
		break;
	    }
	    case io_mul: {
		in_rrr * ir = dynamic_cast<in_rrr *>(op.instr());
		operand op1 = ir->src1_op();
		operand op2 = ir->src2_op();
		if (op1.kind() == OPER_INSTR && 
		    op1.instr()->opcode() == io_ldc &&
		    op2.kind() == OPER_SYM) {
		    in_ldc *ins = dynamic_cast<in_ldc *>(op1.instr());
		    sa->add_var(op2.symbol());
		    h.update_coef(r->get_local(sa->vars[op2.symbol()]), 
				  mult * ins->value().integer());
		} else
		    return false;
		break;
	    }
	    case io_neg: {
		in_rrr * ir = dynamic_cast<in_rrr *>(op.instr());
		if (!fill_q(h, ir->src1_op(), -mult))
		    return false;
		break;
	    }
	    case io_rem: {
		in_rrr * ir = dynamic_cast<in_rrr *>(op.instr());
		operand op1 = ir->src1_op();
		operand op2 = ir->src2_op();
		if (op2.kind() == OPER_INSTR && op2.instr()->opcode() == io_ldc) {
		    in_ldc *ins = dynamic_cast<in_ldc *>(op2.instr());
		    int mod = ins->value().integer(); 
		    Variable_ID q = e->declare();
		    Variable_ID r = e->declare();
		    h.update_coef(r, mult);

		    /* r >= 0 */
		    GEQ_Handle g1 = f->add_GEQ();
		    g1.update_coef(r, 1);

		    /* r <= mod-1 */
		    GEQ_Handle g2 = f->add_GEQ();
		    g2.update_coef(r, -1);
		    g2.update_const(mod-1);

		    /* mod * q + r = op1 */
		    EQ_Handle g3 = f->add_EQ();
		    g3.update_coef(r, -1);
		    g3.update_coef(q, -mod);
		    if (!fill_q(g3, op1, 1))
			return false;
		} else
		    return false;
		break;
	    }
	    default:
		return false;
	}
	break;
    case OPER_SYM:
	sa->add_var(op.symbol());
	h.update_coef(r->get_local(sa->vars[op.symbol()]), mult);
	break;
    default:
	return false;
    }
    return true;
}

boolean guard_constructor::construct_guard(instruction *ins)
{
    switch (ins->opcode()) {
    /* RHS - LHS >= 0 */ 
    case io_sle: { 		    
	in_rrr * ir = dynamic_cast<in_rrr *>(ins);
	GEQ_Handle g = f->add_GEQ();
	if (!fill_q(g, ir->src1_op(), -1))
	    return false;
	if (!fill_q(g, ir->src2_op(), 1))
	    return false;
	break;
    }
    /* RHS - LHS - 1 >= 0 */ 
    case io_sl: { 		    
	in_rrr * ir = dynamic_cast<in_rrr *>(ins);
	GEQ_Handle g = f->add_GEQ();
	g.update_const(-1);
	if (!fill_q(g, ir->src1_op(), -1))
	    return false;
	if (!fill_q(g, ir->src2_op(), 1))
	    return false;
	break;
    }
    /* LHS - RHS == 0 */ 
    case io_seq: { 		    
	in_rrr * ir = dynamic_cast<in_rrr *>(ins);
	EQ_Handle g = f->add_EQ();
	if (!fill_q(g, ir->src1_op(), 1))
	    return false;
	if (!fill_q(g, ir->src2_op(), -1))
	    return false;
	break;
    }
    /* LHS - RHS - 1 >= 0 or RHS - LHS -1 >= 0 */
    case io_sne: {
	in_rrr * ir = dynamic_cast<in_rrr *>(ins);
	F_Or *o = f->add_or();

	GEQ_Handle g1 = o->add_and()->add_GEQ();
	g1.update_const(-1);
	if (!fill_q(g1, ir->src1_op(), -1))
	    return false;
	if (!fill_q(g1, ir->src2_op(), 1))
	    return false;

	GEQ_Handle g2 = o->add_and()->add_GEQ();
	g2.update_const(-1);
	if (!fill_q(g2, ir->src1_op(), 1))
	    return false;
	if (!fill_q(g2, ir->src2_op(), -1))
	    return false;

	break;
    }
    default:
	if (verbose) {
	    fprintf(stderr, "Unhandled guard:\n");
	    ins->print(stderr);
	}
	return false;
    }
    return true;
}

static void set_truth_value(suif_object *o, SRelation *sr,
			    tree_if *ti, tree_node_list *tnl)
{
    if (tnl == ti->then_part()) {
	SRelation *s = (SRelation *)ti->peek_annote(k_if_annote);
	assert(s);
	SRelation *news = sr->Intersection(s);
	delete sr;
	sr = news;
    }
    if (tnl == ti->else_part()) {
	SRelation *s = (SRelation *)ti->peek_annote(k_if_annote);
	assert(s);
	SRelation *news = sr->Difference(s);
	delete sr;
	sr = news;
    }
    SRelation *s = (SRelation *)o->get_annote(k_if_annote);
    if (s) {
	SRelation *news = sr->Union(s);
	delete sr;
	delete s;
	sr = news;
    }
    o->append_annote(k_if_annote, sr);
}

boolean domain_constructor::construct_guard(tree_if *ti, tree_node_list *tnl)
{
    tree_node_list_iter tnli(tnl);

    for (int i = 0; !tnli.is_empty(); ++i) {
	tree_node *tn = tnli.step();
	switch (tn->kind()) {
	case TREE_INSTR: {
	    tree_instr *in = (tree_instr*)tn;
	    switch (in->instr()->opcode()) {
	    case io_btrue:
	    case io_bfalse: {
		in_bj *bj = (in_bj *)in->instr();
		operand op = bj->src_op();
		SRelation *sr = NULL;
		switch (op.kind()) {
		case OPER_INSTR: {
		    instruction * ins = op.instr();
		    guard_constructor gc(sa, verbose);
		    bool ok = gc.construct_guard(ins);
		    if (ok) {
			if (in->instr()->opcode() == io_bfalse)
			    sr = new SRelation(*gc.r, sa);
			else {
			    Relation r = Complement(*gc.r);
			    sr = new SRelation(r, sa);
			}
		    }
		    break;
		}
		case OPER_SYM: {
		    sym_node *v= op.symbol();
		    SRelation *s = (SRelation *)v->peek_annote(k_if_annote);
		    assert(s);
		    if (in->instr()->opcode() == io_bfalse)
			sr = new SRelation(*s);
		    else
			sr = s->Complement();
		    break;
		}
		default:
		    assert(0);
		}
		assert(tnl == ti->header());
		if (sr)
		    ti->append_annote(k_if_annote, sr);
		return sr != NULL;
	    }
	    case io_sl:
	    case io_sle: {
		guard_constructor gc(sa, verbose);
		boolean ok = gc.construct_guard(in->instr());
		assert(in->instr()->dst_op().kind() == OPER_SYM);
		sym_node* v = in->instr()->dst_op().symbol();
		if (ok)
		    set_truth_value(v, new SRelation(*gc.r, sa), ti, tnl);
		else
		    return ok;
		break;
	    }
	    case io_cpy: {
		assert(in->instr()->dst_op().kind() == OPER_SYM);
		sym_node* vd = in->instr()->dst_op().symbol();
		assert(in->instr()->src_op(0).kind() == OPER_SYM);
		sym_node* vs = in->instr()->src_op(0).symbol();
		SRelation *s = (SRelation *)vs->peek_annote(k_if_annote);
		assert(s);
		set_truth_value(vd, new SRelation(*s), ti, tnl);
		break;
	    }
	    case io_mrk:
		break;
	    default:
		if (verbose) {
		    fprintf(stderr, "Unhandled guard:\n");
		    in->instr()->print(stderr);
		}
		return false;
	    }
	    break;
	}
	case TREE_IF: {
	    tree_if *ti = (tree_if *) tn;
	    bool ok = construct_guard(ti, ti->header()) &&
		      construct_guard(ti, ti->then_part()) &&
		      construct_guard(ti, ti->else_part());
	    if (!ok)
		return ok;
	    break;
	}
	default:
	    return false;
	}
    }
    return true;
}

boolean domain_constructor::scan(tree_node * tn)
{
    boolean ok = true;
    stmt *stmt = NULL;

    switch (tn->kind()) {
    case TREE_INSTR: {
	tree_instr *in = (tree_instr*)tn;
	switch (in->instr()->opcode()) {
	case io_memcpy:
	case io_cpy:
	case io_str:
	case io_lod:
	case io_cvt:
	case io_ldc:
	case io_add:
	case io_sub:
	case io_div:
	case io_cal:
	    root = in->instr();
	    stmt = extract_stmt(in);
	    if (stmt)
		root->append_annote(k_stmt, stmt);
	    ok = stmt != NULL;
	    break;
	case io_mrk:
	    break;
	default:
	    if (verbose) {
		fprintf(stderr, "Unhandled instruction:\n");
		in->instr()->print(stderr);
	    }
	    ok = false;
	    break;
	}
	break;
    }
    case TREE_IF: {
	tree_if *ti = (tree_if *) tn;
	ok = construct_guard(ti, ti->header()) && scan(ti->then_part()) 
					       && scan(ti->else_part());
	break;
    }
    case TREE_FOR: {
	dep_for_annote * df = 
		(dep_for_annote *)tn->peek_annote(k_dep_for_annote);
	named_lin_ineq *lb, *ub;
	tree_for * tf = df->my_for;
	sa->add_var(tf->index());
	sa->vars[tf->index()]->iterator = true;
	immed idx(tf->index());
	lb = named_lin_ineq::mk_named_lin_ineq(*df->lb, idx, TRUE);
	ub = named_lin_ineq::mk_named_lin_ineq(*df->ub, idx, FALSE);
	ok = lb && ub;
	if (lb)
	    delete lb;
	if (ub)
	    delete ub;
	ok = scan(tf->body());
	break;
    }
    case TREE_LOOP: {
	/* We only accept infinite loops */
	tree_loop *tl = static_cast<tree_loop *>(tn);
	tree_node_list *tnl = tl->test();
	tree_node_list_iter tnli(tnl);
	bool infinite_loop = false;

	/* Simple test for infinite loop:
	 * a loop whose test jumps right back to the top of the loop.
	 */
	while (!tnli.is_empty()) {
	    tree_node *t = tnli.step();
	    if (t->kind() != TREE_INSTR)
		break;
	    tree_instr *in = static_cast<tree_instr*>(t);
	    if (in->instr()->opcode() == io_mrk)
		continue;
	    if (in->instr()->opcode() != io_jmp)
		break;
	    in_bj *bj = static_cast<in_bj *>(in->instr());
	    if (bj->target() != tl->toplab())
		break;
	    infinite_loop = true;
	    break;
	}
	ok = infinite_loop && scan(tl->body());
	break;
    }
    default:
	unsigned n = tn->num_child_lists();
	for (unsigned i = 0; ok && i < n; ++i)
	    ok = scan(tn->child_list_num(i));
    }

    return ok;
}

boolean domain_constructor::scan(tree_node_list * tnl)
{
    tree_node_list_e *start = NULL;

    tree_node_list_iter tnli(tnl);
    for (int i = 0; !tnli.is_empty(); ++i) {
	tree_node *t = tnli.step();
	if (scan(t)) {
	    if (!start)
		start = t->list_e();
	} else if (start) {
	    tree_node *root;

	    if (start->next() == t->list_e())
		root = start->contents;
	    else {
		block_symtab *st = new block_symtab("static root");
		tnl->scope()->add_child(st);
		tree_node_list *tl = new tree_node_list;
		tree_block *tb = new tree_block(tl, st);
		tnl->insert_before(tb, start);
		tb->proc()->block()->number_instrs();

		tree_node_list_e *next = start->next();
		while (start != t->list_e()) {
		    tnl->remove(start);
		    tl->append(start);
		    start = next;
		    next = start->next();
		}
		root = tb;
	    }
	    if (!(root->kind() == TREE_INSTR && 
		  ((tree_instr*)root)->instr()->opcode() == io_mrk))
		root->append_annote(k_static_root, new immed_list());

	    start = NULL;
	}
    }

    return start == tnl->head();
}

void domain_constructor::scan(context_info *ctx)
{
    sa = new symtab_ann;
    sa->symtab = new global_symtab("statement iterators");
    if (ctx) {
	std::map<var_sym *, Param_Iterator *>::iterator gi;
	for (gi = ctx->sa->vars.begin(); gi != ctx->sa->vars.end(); ++gi)
	    sa->vars[(*gi).first] = (*gi).second;
    }
    tp->append_annote(k_domain_symtab, sa);
    scan(tp->body());
}

void construct_domains(tree_proc * tp, int verbose, int oldstyle, context_info *ctx)
{
    domain_constructor dc(tp, verbose, oldstyle);
    dc.scan(ctx);
    dc.construct();
}

static void free_stmt(void *data)
{
    stmt *s = (stmt *)data;
    delete s;
}

static void
free_domain_ann(void *data)
{
    domain_ann *da = (domain_ann *) data;
    delete da->domain;
    delete da;
}

static void
free_symtab(void *data)
{
    symtab_ann *sa = (symtab_ann *) data;
    delete sa->symtab;
    delete sa;
}

static void free_srelation(void *data)
{
    SRelation *sr = (SRelation *) data;
    delete sr;
}

void
init(int &argc, char *argv[])
{
    ANNOTE(k_static_root, "pers_static_root", FALSE);
    STRUCT_ANNOTE(k_stmt, "pers_stmt", FALSE,
	NULL, NULL, free_stmt, NULL);
    STRUCT_ANNOTE(k_iteration_domain, "pers_iteration_domain", FALSE,
	NULL, NULL, free_domain_ann, NULL);
    STRUCT_ANNOTE(k_domain_symtab, "pers_domain_symtab", FALSE,
	NULL, NULL, free_symtab, NULL);
    STRUCT_ANNOTE(k_if_annote, "pers_if_annote", FALSE,
	NULL, NULL, free_srelation, NULL);
}
