#include "suif_inc.h"
#include <sstream>
#include "domain.h"

#include "yaml.h"
#include "pdg.h"

#define ARRAY_SIZE(a) (sizeof(a)/sizeof(*a))

using pdg::PDG;

static PDG *main_pdg;
static char *pdg_name = NULL;

class instrumenter {
    tree_node* add_iters(tree_node_list* root, tree_node* tn, block& iter, block& g);
    void add_iters(tree_node_list* root, tree_node_list* tnl, block& iter, block& g);
public:
    instrumenter(tree_proc *t, PDG* p, char * n);

    void instrument(tree_node *root);
    void visit(tree_node *t);

    tree_proc 	   *tp;
    PDG	   	   *pdg;
    global_symtab  *symtab;
    var_sym_v		its;
    int			level;
    int			forlevel;
    map<int,pdg::node*>	nodes;
    proc_sym 		    *access;
    proc_sym 		    *set_name;
    char		    *filename;
};

/*
class my_block : public block {
public:
    block* add_call(block& s, const block & in) {
	return add_call(s.child, in);
    }
};
*/

static void visit (tree_node *t, void *d)
{
    ((instrumenter *)d)->visit(t);
}

static void do_proc(tree_proc * tp)
{
    proc_sym * psym = tp->proc();
    fill_in_access(tp);
    construct_domains(tp, 0, 0, NULL);
    instrumenter i(tp, main_pdg, pdg_name);
    tp->map(::visit, (void *) &i);
}

int main(int argc, char * argv[])
{
    static cmd_line_option my_options[] = {
	{ CLO_STRING,   "-pdg",		"", 	&pdg_name },
    };
    
    start_suif(argc, argv);
    parse_cmd_line(argc, argv, my_options, ARRAY_SIZE(my_options));
    init(argc, argv);

    assert(pdg_name);
    FILE *input = fopen(pdg_name, "r");
    assert(input);

    main_pdg = yaml::Load<PDG>(input);

    fclose(input);

    suif_proc_iter(argc, argv, do_proc, TRUE);

    if (pdg_name)
	delete [] pdg_name;

    return 0;
}

tree_node* instrumenter::add_iters(tree_node_list* root,
				   tree_node* tn, block& iter, block& g)
{
    tree_node* last = tn;

    switch (tn->kind()) {
    case TREE_INSTR: {
	tree_instr *in = (tree_instr*)tn;
	pdg::node *s = nodes[in->instr()->number()];
	if (!s)
	    break;

	int f, k;
	for (k = 0, f = 0; k < s->prefix.size(); ++k) {
	    if (s->prefix[k] == -1) {
		block n(block::CALL(set_name, g, s->nr, f, 
					block(its[k]->name())));
		root->insert_after(n.make_tree_node(), 0);
		++f;
	    }
	}

	for (int i = 0; i < s->statement->accesses.size(); ++i) {
	    pdg::UnionMap *map = s->statement->accesses[i]->array->map;
	    assert(map->constraints.size() == 1);
	    symtab_ann *sa = new symtab_ann();
	    SRelation *alloc = new SRelation(map->input, map->output,
					     map->constraints[0]->el, sa);
	    map = s->statement->accesses[i]->map;
	    assert(map->constraints.size() == 1);
	    SRelation *acc = new SRelation(map->input, map->output,
					      map->constraints[0]->el, sa);
	    SRelation *comp = alloc->Composition(acc);
	    std::vector<std::vector<std::vector<int> > > c = comp->constraints();
	    assert(c.size() == 1);
	    assert(c[0].size() == 1);

	    int l = c[0][0].size();
	    int factor = -c[0][0][l-2];

	    block e(factor * c[0][0][l-1]);
	    for (int k = 0; k < l - 3; ++k)
		if (c[0][0][k+1] != 0)
		    e.set(e + (block(factor * c[0][0][k+1]) * 
			       block(its[k])));

	    block a(block::CALL(access, g, s->nr, i, 
				e / block(pdg->cacheline)));
	    last = last->parent()->insert_after(a.make_tree_node(),
						last->list_e());
	}
	break;
    }
    case TREE_FOR: {
	dep_for_annote * df = 
		(dep_for_annote *)tn->peek_annote(k_dep_for_annote);
	tree_for * tf = df->my_for;

	if (level >= its.size())
	    its.push_back(tf->index());
	else
	    its[level] = tf->index();

	++level;
	++forlevel;
	add_iters(root, tf->body(), iter, g);
	--forlevel;
	--level;

	block set(iter[block(forlevel)] = block(tf->index()));
	tf->body()->insert_after(set.make_tree_node(), 0);

	break;
    }
    case TREE_IF: {
	tree_if * ti = (tree_if *)tn;

	add_iters(root, ti->then_part(), iter, g);

	if (!ti->else_part()->is_empty())
	    add_iters(root, ti->else_part(), iter, g);
	break;
    }
    }
    return last;
}

void instrumenter::add_iters(tree_node_list *root,
			    tree_node_list* tnl, block& iter, block& g)
{
    if (level >= its.size() || its[level]->parent() != symtab)
	its.push_back(symtab->new_unique_var(type_signed));

    tree_node_list_iter tnli(tnl);
    for (int i = 0; !tnli.is_empty(); ++i) {
	tree_node *t = tnli.step();

	++level;
	tree_node *last = add_iters(root, t, iter, g);
	--level;

	/*
	printf("current: %p, last: %p\n", t, last);
	while (t != last) {
	    t = tnli.step();
	    printf("current: %p, last: %p\n", t, last);
	}
	*/
    }
}

instrumenter::instrumenter(tree_proc *t, PDG* p, char *n) 
	: tp(t), pdg(p), filename(n) {
    symtab = new global_symtab("statement iterators");
    for (int i = 0; i < pdg->nodes.size(); ++i) {
	pdg::node* s = pdg->nodes[i];
	int op = s->statement->operation;
	assert(nodes.find(op) == nodes.end());
	nodes[op] = s;
    }

    global_symtab *globals = tp->proc()->file()->parent()->globals();
    type_node *temp_type;
    temp_type = block::parse_type(globals, "void (void *, int, int, long)");
    assert(temp_type->is_func());
    access = globals->new_proc( (func_type*)temp_type, src_c, 
				"reuse_check_access");
}

void instrumenter::visit(tree_node *t)
{
    if (!t->peek_annote(k_static_root))
	return;
    if (t->number() != pdg->root)
	return;
    instrument(t);
}

void instrumenter::instrument(tree_node *r)
{
    tree_block *root = dynamic_cast<tree_block *>(r);
    global_symtab *globals = tp->proc()->file()->parent()->globals();
    block::set_proc(tp);

    type_node *temp_type;
    temp_type = block::parse_type(globals, "void * (char *, int *)");
    assert(temp_type->is_func());
    proc_sym *init = globals->new_proc( (func_type*)temp_type, src_c, 
					"reuse_check_init");

    temp_type = block::parse_type(globals, "void (void *, int, int, char*)");
    assert(temp_type->is_func());
    set_name = globals->new_proc( (func_type*)temp_type, src_c, 
					    "reuse_check_set_name");

    temp_type = block::parse_type(globals, "void (void *)");
    assert(temp_type->is_func());
    proc_sym *end = globals->new_proc( (func_type*)temp_type, src_c, 
					"reuse_check_end");

    tree_node_list *tnl = ((tree_block*) root)->body();
    block pdgb(block::new_sym(block::parse_type("void *"), "pdg", 
			     root->symtab()));
    std::ostringstream os;
    os << "int [" << pdg->dimension << "]";
    block iter(block::new_sym(block::parse_type(os.str().c_str()), "iter", 
			     root->symtab()));

    level = 0;
    forlevel = 0;
    add_iters(root->body(), root->body(), iter, pdgb);

    block name(filename);
    //block s(pdgb = block::CALL(init, name, block::op(uop_addr, iter)));
    //block s(pdgb = block::CALL(init, name, iter.addr()));
    block s(pdgb = block::CALL(init, name, iter));
    block e(block::CALL(end, pdgb));
    tnl->insert_after(s.make_tree_node(), 0);
    tnl->append(e.make_tree_node());
}
