#include <alloca.h>
#include <getopt.h>
#include <set>
#include <sstream>
#include <libxml/xmlwriter.h>

#include "yaml.h"
#include "pdg.h"
#include "pdgutil.h"
extern "C" {
#include "piputil.h"
}
#include "cloog_util.h"
#include "xml_AST.h"

struct option options[] = {
    { "input",	required_argument,  0,  'i' },
    { "output",	required_argument,  0,  'o' },
    { 0, 0, 0, 0 }
};

using namespace std;
using pdg::PDG;

static Polyhedron *constraints2polyhedron(int dim, 
					  std::vector< std::vector<int> > & c)
{
    if (c.size() == 0)
	return Universe_Polyhedron(dim);

    Matrix *M = Matrix_Alloc(c.size(), c[0].size());
    for (int i = 0; i < c.size(); ++i) {
	for (int j = 0; j < c[i].size(); ++j) {
	    value_set_si(M->p[i][j], c[i][j]);
	}
    }
    Polyhedron *P = Constraints2Polyhedron(M, POL_NO_DUAL);
    Matrix_Free(M);
    return P;
}

static Matrix *extract_map(Polyhedron *map, int dim)
{
    Matrix *M;
    int i;
    Value mone;
    assert(map->NbConstraints >= dim);
    value_init(mone);
    value_set_si(mone, -1);
    M = Matrix_Alloc(dim, 1 + map->Dimension - dim);
    for (i = 0; i < dim; ++i) {
	assert(value_zero_p(map->Constraint[i][0]));
	assert(value_one_p(map->Constraint[i][1+i]));
	Vector_Scale(map->Constraint[i]+1+dim, M->p[i], mone, 
		     1 + map->Dimension - dim);
    }
    value_clear(mone);
    return M;
}

/* Dependency relations are usually specified as a "dependent
 * mapping", where the input variables are given as an integer
 * linear combination of other variables.
 * However, we sometimes use Omega to manipulate these relations
 * and then Omega may simplify way some existentially quantified
 * variables, e.g., { [i,o] | exists a : i = a, o = 2a } may
 * be simplified to { [i,o] | o = 2i } where i is no longer
 * an integer linear combination of other variables.
 *
 * In this function we (re)introduce existentially quantified variables
 * for each input variable that is not an integer linear combination
 * of other variables.  In particular, if i is not an integer linear
 * combination of other variables, we introduce a new variable a
 * right before the parameters that is set equal to i and introduce an extra
 * equality i = a.
 */
static Polyhedron *ensure_dependent_mapping(Polyhedron *P,
			unsigned in, unsigned out, unsigned nparam)
{
    Matrix *M;
    Polyhedron *Q;
    int i, j;
    int n;
    unsigned extra;
    int *pos = (int *)alloca(in * sizeof(int));

    assert(in + out + nparam <= P->Dimension);
    extra = P->Dimension - (in + out + nparam);

    for (i = 0, n = 0; i < in; ++i) {
	assert(value_zero_p(P->Constraint[i][0]));
	assert(value_notzero_p(P->Constraint[i][1+i]));
	if (value_one_p(P->Constraint[i][1+i]))
	    pos[i] = i;
	else {
	    pos[i] = in + out + extra + n;
	    ++n;
	}
    }
    if (!n)
	return P;

    M = Matrix_Alloc(P->NbConstraints + n, 1 + P->Dimension + n + 1);
    /* Introduce i = a constraints for each new variable */
    for (i = 0; i < in; ++i) {
	if (pos[i] == i)
	    continue;
	j = pos[i] - (in + out + extra);
	value_set_si(M->p[j][1+j], 1);
	value_set_si(M->p[j][1+pos[i]], -1);
    }

    /* Copy original constraints, moving the column for variable
     * i to that of variable a.  This is a minor optimization.
     * Constraints2Polyhedron would perform the same transformation
     * if the constraints would simply be copied.
     */
    for (i = 0; i < P->NbConstraints; ++i) {
	value_assign(M->p[n+i][0], P->Constraint[i][0]);
	for (j = 0; j < in; ++j)
	    value_assign(M->p[n+i][1+pos[j]], P->Constraint[i][1+j]);
	Vector_Copy(P->Constraint[i]+1+in, M->p[n+i]+1+in, out+extra);
	Vector_Copy(P->Constraint[i]+1+in+out+extra,
		    M->p[n+i]+1+in+out+extra+n, nparam+1);
    }
    Polyhedron_Free(P);
    Q = Constraints2Polyhedron(M, POL_NO_DUAL);
    Matrix_Free(M);
    return Q;
}

struct espam_edge {
    char	  *name;
    char	  *from_port;
    char	  *to_port;
    pdg::UnionSet *from_domain;
    pdg::UnionSet *to_domain;
    pdg::node    *from_node;
    pdg::node    *to_node;
    vector<pdg::access *>    from_access;
    vector<pdg::access *>    to_access;
    pdg::Matrix  *map;
    pdg::array *array;
    int		reordering;
    int		multiplicity;
    integer *	size;
    int		nr;
    bool	sticky;
    bool	shift_register;

    espam_edge() : from_domain(NULL), to_domain(NULL), from_node(NULL), to_node(NULL),
	     map(NULL), array(NULL), reordering(-1), multiplicity(-1), size(NULL) {}
};

/* Returns the number of control variables in a domain.
 * If the node_domain is a union, we just hope there are no control variabels
 * for now.
 */
static unsigned domain_n_control(pdg::UnionSet *node_domain)
{
    unsigned dim = node_domain->dim;
    unsigned nparam = node_domain->params.size();
    if (node_domain->constraints.size() > 1) {
	for (int i = 0; i < node_domain->constraints.size(); ++i) {
	    assert(node_domain->constraints[i]->el.size() >= 1);
	    int nexist = node_domain->constraints[i]->el[0].size() - 2 - dim - nparam;
	    assert(nexist == 0);
	}
	return 0;
    }
    assert(node_domain->constraints.size() == 1);
    assert(node_domain->constraints[0]->el.size() >= 1);
    return node_domain->constraints[0]->el[0].size() - 2 - dim - nparam;
}

/* The ESPAM people want all "control" variables of the node domain
 * to appear in the port domains.
 * Moreover, the variables in the (reverse) "mapping"s should correspond
 * to the variables in the port domains.  We therefore need to perform
 * the transformation on the "to" domain before extracting the mapping.
 * (Recall that a dependence is a relation from the source of the data
 * to the user of the data, while the mapping in the ESPAM data structures
 * maps iterations of an input port to iterations of an output port.)
 * The "from" domain is handled later in normalize_domain.
 */
static Polyhedron *normalize_to_domain(PDG *pdg, Polyhedron *P, unsigned in,
				       pdg::UnionSet *to_node_domain,
				       const std::vector<Global_Var_Decl *>& params)
{
    unsigned out = to_node_domain->dim;
    unsigned nparam = to_node_domain->params.size();
    unsigned nexist = domain_n_control(to_node_domain);

    if (!nexist)
	return P;

    to_node_domain->dim += nexist;
    Relation domain = pdg::set2relation(to_node_domain, params);
    to_node_domain->dim -= nexist;
    pdg::UnionMap *map = new pdg::UnionMap(in, out, P, pdg->params);
    Polyhedron_Free(P);
    Relation mr = pdg::map2relation(map, params);
    delete map;
    mr = Extend_Range(mr, nexist);
    mr = Restrict_Range(mr, domain);
    mr.simplify();
    map = new pdg::OUnionMap(&mr, params);
    map->output -= nexist;
    P = *map;
    delete map;
    return P;
}

static void add_edge(PDG *pdg, vector<espam_edge*>& split_edges,
		     pdg::dependence *dep, pdg::UnionMap *rel,
		     pdg::UnionSet *to_node_domain,
		     const vector<pdg::access *>& to_accesses, bool sticky,
		     int i, const std::vector<Global_Var_Decl *> &params)
{
    int nparam = pdg->params.size();
    char buf[10];
    snprintf(buf, 10, "ED_%d", i);

    for (int k = 0; k < rel->constraints.size(); ++k) {
	espam_edge *edge = new espam_edge;
	int dim = rel->input + rel->output;
	Polyhedron *P = constraints2polyhedron(dim, rel->constraints[k]->el);
	if (dep->multiplicity && !dep->reordering) {
	    /* ESPAM expects the input port of a BroadcastInOrder channel
	     * to only consist of the iterations that first read a value
	     * because that is what Compaan gives.
	     */
	    Polyhedron *P2 = pip_lexminmax(P, rel->input, 
				P->Dimension-rel->input-nparam, nparam, 0);
	    Polyhedron_Free(P);
	    P = P2;
	    assert(!P->next);
	}
	P = normalize_to_domain(pdg, P, rel->input, to_node_domain, params);
	assert(P);
	P = ensure_dependent_mapping(P, rel->input, rel->output, nparam);
	Polyhedron *domain = Polyhedron_projectout(P, rel->input, 
					P->Dimension-rel->input-nparam, nparam);
	Polyhedron *range = Polyhedron_projectout_dep(P, rel->input);
	Matrix *M = extract_map(P, rel->input);
	Polyhedron_Free(P);
	edge->reordering = dep->reordering;
	edge->multiplicity = dep->multiplicity;
	edge->size = dep->value_size;
	edge->array = dep->array;
	edge->from_node = dep->from;
	edge->to_node = dep->to;
	edge->from_access.push_back(dep->from_access);
	edge->to_access = to_accesses;
	edge->from_domain = new pdg::UnionSet(rel->input, domain, pdg->params);
	edge->to_domain = new pdg::UnionSet(rel->output, range, pdg->params);
	edge->map = new pdg::Matrix(M);
	edge->nr = i;
	edge->name = strdup(buf);
	edge->sticky = sticky;
	edge->shift_register = dep->type == pdg::dependence::pn_shift;
	Matrix_Free(M);
	Domain_Free(domain);
	Domain_Free(range);
	split_edges.push_back(edge);
    }
}

static void writeMatrix(xmlTextWriterPtr writer, const char *name, pdg::Matrix *M)
{
    int rc;
    std::ostringstream strm;

    rc = xmlTextWriterStartElement(writer, BAD_CAST name);
    assert(rc >= 0);

    strm << "[";
    for (int i = 0; i < M->el.size(); ++i) {
	if (i != 0)
	    strm << "; ";
	for (int j = 0; j < M->el[i].size(); ++j) {
	    if (j != 0)
		strm << ", ";
	    strm << M->el[i][j];
	}
    }
    strm << "]";

    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "matrix", 
				     BAD_CAST strm.str().c_str());
    assert(rc >= 0);

    rc = xmlTextWriterEndElement(writer);
    assert(rc >= 0);
}

struct div_info {
    string s;
    vector<int> coeff;
    int denom;
};

struct div_vector {
    int		      index;
    vector<div_info*> divs;
};

/* For constraint
 *		<a, i> + -m e + <b, p> >= 0
 * check if there is a div equal to
 *		e = floor((<a,i> + <b,p>)/m) = div(<a,i> + <b,p>, m)
 * If not, add such a div to divs.
 * The index of the variable in divs is returned.
 */
int div_index(vector<div_info*> &divs, std::vector<int> &constraint,
	      unsigned pos, unsigned dim, unsigned nexist,
	      const seq<pdg::parameter>& params)
{
    int j;
    std::ostringstream strm;
    unsigned nparam = params.size();
    div_info *di = new div_info;
    di->denom = -constraint[1+pos];
    int s = 1;
    if (di->denom < 0) {
	di->denom = -di->denom;
	s = -1;
    }
    for (j = 0; j < dim; ++j)
	di->coeff.push_back(s * constraint[1+j]);
    for (j = 0; j < nparam; ++j)
	di->coeff.push_back(s * constraint[1+dim+nexist+j]);
    di->coeff.push_back(s * constraint[1+dim+nexist+nparam]);

    for (j = 0; j < divs.size(); ++j) {
	if (divs[j]->denom != di->denom || divs[j]->coeff != di->coeff)
	    continue;
	delete di;
	return j;
    }

    divs.push_back(di);
    strm << "div(";
    bool first = true;
    for (j = 0; j < dim; ++j) {
	if (di->coeff[j] == 0)
	    continue;
	if (!first && di->coeff[j] >= 0)
	    strm << "+";
	first = false;
	strm << di->coeff[j] << "*" << "c" << j;
    }
    for (j = 0; j < nparam; ++j) {
	if (di->coeff[dim+j] == 0)
	    continue;
	if (!first && di->coeff[dim+j] >= 0)
	    strm << "+";
	first = false;
	strm << di->coeff[dim+j] << "*" << params[j]->name->s;
    }
    if (di->coeff[dim+nparam] != 0) {
	if (!first && di->coeff[dim+nparam] >= 0)
	    strm << "+";
	strm << di->coeff[dim+nparam];
    }
    strm << "," << di->denom << ")";
    di->s = strm.str();
    return divs.size()-1;
}

int only_set(std::vector<int> &constraint, unsigned pos, unsigned n,
		unsigned only)
{
    for (int j = 0; j < n; ++j) {
	if (only == pos+j)
	    continue;
	if (constraint[1+pos+j] != 0)
	    return 0;
    }
    return 1;
}

/* Check whether the variable at position pos is the greatest integer
 * part of an affine expression in the iterators and parameters.
 * If so, it is added to divs, unless div already contains an equivalent
 * variable.
 * The index of the variable in divs is returned.
 *
 * We currently only check for divs described by constraints of the
 * following form
 *
 *		<a, i> + -m e + <b, p> = 0
 * i.e.,
 *		e = floor((<a,i> + <b,p>)/m) = div(<a,i> + <b,p>, m)
 *
 * or a pair of constraints
 *		 <a, i> + -m e +  <b, p> + k1 >= 0
 *		<-a, i> +  m e + <-b, p> + k2 >= 0
 * with 0 <= k1 + k2 < m
 * The corresponding div is then
 *		e = floor((<a,i> + <b,p> + k1)/m) = div(<a,i> + <b,p> + k1, m)
 */
static int find_div_index(vector<div_info*>& divs, int pos,
			  pdg::Matrix *constraints, int dim, 
			  const seq<pdg::parameter>& params)
{
    assert(pos >= dim);
    int nparam = params.size();
    int nexist = constraints->el[0].size() - 2 - dim - nparam;
    int i;
    for (i = 0; i < constraints->el.size(); ++i) {
	if (constraints->el[i][0] != 0)
	    continue;
	if (constraints->el[i][1+pos] == 0)
	    continue;
	if (!only_set(constraints->el[i], dim, nexist, pos))
	    continue;
	return div_index(divs, constraints->el[i], pos, dim, nexist, params);
    }
    for (i = 0; i < constraints->el.size(); ++i) {
	if (constraints->el[i][0] != 1)
	    continue;
	if (constraints->el[i][1+pos] == 0)
	    continue;
	if (!only_set(constraints->el[i], dim, nexist, pos))
	    continue;
	int m = constraints->el[i][1+pos];
	if (m < 0)
	    m = -m;
	for (int j = i+1; j < constraints->el.size(); ++j) {
	    if (constraints->el[j][0] != 1)
		continue;
	    if (constraints->el[j][1+pos] != -constraints->el[i][1+pos])
		continue;
	    if (!only_set(constraints->el[j], dim, nexist, pos))
		continue;
	    int k;
	    k = constraints->el[i][1+dim+nexist+nparam] +
		constraints->el[j][1+dim+nexist+nparam];
	    if (k < 0 || k >= m)
		continue;
	    for (k = 0; k < dim; ++k)
		if (constraints->el[j][1+k] != -constraints->el[i][1+k])
		    break;
	    if (k < dim)
		continue;
	    for (k = 0; k < nparam; ++k)
		if (constraints->el[j][1+dim+nexist+k] !=
			-constraints->el[i][1+dim+nexist+k])
		    break;
	    if (k < nparam)
		continue;
	    if (constraints->el[i][1+pos] > 0)
		i = j;
	    return div_index(divs, constraints->el[i],
			     pos, dim, nexist, params);
	}
    }
    /* Gaussian elimination may have made the existential variable
     * unrecognizable.  Put it last (if it isn't already) and
     * redo the elimination.
     */
    if (pos != dim+nexist-1) {
	int eq;
	int index;
	Matrix *M = *constraints;
	ExchangeColumns(M, 1+pos, 1+dim+nexist-1);
	for (eq = 0; eq < M->NbRows; ++eq)
	    if (value_notzero_p(M->p[eq][0]))
		break;
	Gauss(M, eq, M->NbColumns-1);
	constraints = new pdg::Matrix(M);
	Matrix_Free(M);
	index = find_div_index(divs, dim+nexist-1, constraints, dim, params);
	delete constraints;
	return index;
    }
    assert(0);
}

static unsigned n_statement_dims(pdg::PDG *pdg)
{
    unsigned nstat = 0;
    for (int i = 0; i < pdg->statement_dimensions.size(); ++i)
	if (pdg->statement_dimensions[i])
	    ++nstat;
    return nstat;
}

/* Remove the columns from M that refer to statement level dimensions
 * and remove the rows that have become zero by this removal.
 * As an added bonus, we also remove the "positivity constraint"
 * that PolyLib may have added.
 */
static pdg::Matrix *strip_statement_dims(pdg::PDG *pdg, pdg::Matrix *M)
{
    pdg::Matrix *stripped = new pdg::Matrix;
    int dim = pdg->statement_dimensions.size();
    unsigned nstat = n_statement_dims(pdg);
    for (int i = 0; i < M->el.size(); ++i) {
	vector<int> row(M->el[i].size() - nstat);
	bool nonzero = false;
	bool stat = false;
	for (int j = 0, k = 0; j < M->el[i].size(); ++j) {
	    if (j >= 1 && j <= M->el[i].size()-2 && M->el[i][j] != 0) {
		if (j <= dim && pdg->statement_dimensions[j-1])
		    stat = true;
		else
		    nonzero = true;
	    }
	    if (1 <= j && j <= dim && pdg->statement_dimensions[j-1])
		continue;
	    row[k++] = M->el[i][j];
	}
	assert(!(nonzero && stat));
	if (nonzero)
	    stripped->el.push_back(row);
    }
    return stripped;
}

/* Remove the rows and columns from M that refer to statement level dimensions.
 */
static pdg::Matrix *strip_statement_dims_from_map(pdg::PDG *pdg, pdg::Matrix *M)
{
    pdg::Matrix *stripped = new pdg::Matrix;
    unsigned dim = pdg->statement_dimensions.size();
    unsigned nstat = n_statement_dims(pdg);

    for (int i = 0; i < M->el.size(); ++i) {
	if (pdg->statement_dimensions[i])
	    continue;

	vector<int> row(M->el[i].size() - nstat);
	for (int j = 0, k = 0; j < M->el[i].size(); ++j) {
	    if (j < dim && pdg->statement_dimensions[j])
		continue;
	    row[k++] = M->el[i][j];
	}
	stripped->el.push_back(row);
    }
    return stripped;
}

/* "control" variables are first looked up in parent_domain and
 * then in domain.
 */
static void writeLinearBound(xmlTextWriterPtr writer, pdg::UnionSet *domain, 
			     pdg::UnionSet *parent_domain,
			     pdg::PDG *pdg, div_vector& dv, int part)
{
    int rc;
    std::ostringstream strm;
    set<int> controls;

    rc = xmlTextWriterStartElement(writer, BAD_CAST "linearbound");
    assert(rc >= 0);

    for (int i = 0, j = 0; i < domain->dim; ++i) {
	if (pdg->statement_dimensions[i])
	    continue;
	if (j++ != 0)
	    strm << ", ";
	strm << "c" << i;
    }

    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "index", 
				     BAD_CAST strm.str().c_str());
    assert(rc >= 0);

    int parent_nexist = 0;
    /* if parent_domain is a union, we assume that it has no control
     * variables.  Otherwise normalize_domain would have complained.
     */
    if (parent_domain->constraints.size() == 1 &&
	    parent_domain->constraints[0]->el.size() > 0)
	parent_nexist = parent_domain->constraints[0]->el[0].size() - 2 
			    - parent_domain->dim - parent_domain->params.size();

    int nexist = 0;
    if (domain->constraints[part]->el.size() > 0)
	nexist = domain->constraints[part]->el[0].size() - 2 
			- domain->dim - domain->params.size();

    strm.str("");
    for (int i = 0; i < parent_nexist; ++i) {
	int pos = find_div_index(dv.divs, parent_domain->dim+i,
				 parent_domain->constraints[0],
				 parent_domain->dim, parent_domain->params);
	if (i != 0)
	    strm << ", ";
	strm << "e" << pos+dv.index;
	controls.insert(pos);
    }
    for (int i = parent_nexist; i < nexist; ++i) {
	int pos = find_div_index(dv.divs, domain->dim+i, domain->constraints[part],
				 domain->dim, domain->params);
	if (i != 0)
	    strm << ", ";
	strm << "e" << pos+dv.index;
	controls.insert(pos);
    }

    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "staticControl", 
				     BAD_CAST strm.str().c_str());
    assert(rc >= 0);

    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "dynamicControl", 
				     BAD_CAST "");
    assert(rc >= 0);

    strm.str("");
    for (int i = 0; i < domain->params.size(); ++i) {
	if (i != 0)
	    strm << ", ";
	strm << domain->params[i]->name->s;
    }

    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "parameter", 
				     BAD_CAST strm.str().c_str());
    assert(rc >= 0);

    pdg::Matrix *stripped = strip_statement_dims(pdg, domain->constraints[part]);
    writeMatrix(writer, "constraint", stripped);
    delete stripped;

    if (pdg->context) {
	assert(pdg->context->constraints.size() == 1);
	writeMatrix(writer, "context", pdg->context->constraints[0]);
    }

    for (set<int>::iterator i = controls.begin(); i != controls.end(); ++i) {
	char buf[10];
	rc = xmlTextWriterStartElement(writer, BAD_CAST "control");
	assert(rc >= 0);

	snprintf(buf, 10, "e%d", *i+dv.index);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name", BAD_CAST buf);
	assert(rc >= 0);

	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "exp", 
					 BAD_CAST dv.divs[*i]->s.c_str());
	assert(rc >= 0);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);
    }

    rc = xmlTextWriterEndElement(writer);
    assert(rc >= 0);
}

static void writeLBS(xmlTextWriterPtr writer, pdg::UnionSet *domain, 
		     pdg::UnionSet *parent_domain,
		     pdg::PDG *pdg, div_vector& dv)
{
    int rc;

    rc = xmlTextWriterStartElement(writer, BAD_CAST "domain");
    assert(rc >= 0);

    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "type", BAD_CAST "LBS");
    assert(rc >= 0);

    assert(domain->constraints.size() >= 1);
    for (int i = 0; i < domain->constraints.size(); ++i)
	writeLinearBound(writer, domain, parent_domain, pdg, dv, i);

    rc = xmlTextWriterEndElement(writer);
    assert(rc >= 0);
}

static char *writePort(xmlTextWriterPtr writer, bool out, int index, char *node_name,
		    const vector<pdg::access *>& accesses, espam_edge* edge,
		    pdg::UnionSet *domain, pdg::UnionSet *parent_domain,
		    pdg::PDG *pdg, div_vector& dv)
{
    char buf[30];
    std::ostringstream strm;
    int rc;
    char *port_name;

    rc = xmlTextWriterStartElement(writer, BAD_CAST (out ? "outport" : "inport"));
    assert(rc >= 0);

    strm << node_name << (out ? 'O' : 'I') << "P_" << edge->name << '_' << index;
    strm << "_V";
    for (int i = 0; i < accesses.size(); ++i)
	strm << '_' << accesses[i]->nr;
    port_name = strdup(strm.str().c_str());

    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name", BAD_CAST port_name);
    assert(rc >= 0);

    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "node", BAD_CAST node_name);
    assert(rc >= 0);

    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "edge", BAD_CAST edge->name);
    assert(rc >= 0);

    for (int i = 0; i < accesses.size(); ++i) {
	rc = xmlTextWriterStartElement(writer, BAD_CAST "bindvariable");
	assert(rc >= 0);

	int nr = accesses[i]->nr;
	bool is_write = accesses[i]->type == pdg::access::write;
	snprintf(buf, 10, "%s_%d", is_write ? "out" : "in", nr);

	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name", BAD_CAST buf);
	assert(rc >= 0);

	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "dataType", 
				     BAD_CAST edge->array->element_type->s.c_str());
	assert(rc >= 0);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);
    }

    writeLBS(writer, domain, parent_domain, pdg, dv);

    rc = xmlTextWriterEndElement(writer);
    assert(rc >= 0);

    return port_name;
}

struct disjoint_edge {
    Relation		    domain;
    vector<pdg::access *>   accesses;
    bool		    sticky;

    disjoint_edge(Relation d, vector<pdg::access *> a, bool s) : 
	domain(d), accesses(a), sticky(s) {}
};

static void disjoin_dependences(vector<disjoint_edge>& parts, 
				const std::vector<Global_Var_Decl *>& globals,
				Relation& remainder, pdg::access *access,
				bool sticky)
{
    int parts_size = parts.size();
    for (int k = 0; k < parts_size; ++k) {
	Relation inter = Intersection(copy(remainder), copy(parts[k].domain));
	if (!inter.is_satisfiable())
	    continue;
	parts[k].domain = Difference(parts[k].domain, copy(inter));
	remainder = Difference(remainder, copy(inter));
	vector<pdg::access *> accesses = parts[k].accesses;
	if (access)
	    accesses.push_back(access);
	parts.push_back(disjoint_edge(inter, accesses, parts[k].sticky || sticky));
    }
    if (remainder.is_satisfiable()) {
	vector<pdg::access *> accesses;
	if (access)
	    accesses.push_back(access);
	parts.push_back(disjoint_edge(remainder, accesses, sticky));
    }
}

/* Scatter dependence and convert to Omega Relation
 */
static Relation scat_dep_relation(pdg::dependence *dep,
				  const std::vector<Global_Var_Decl *>& params)
{
    pdg::UnionMap *rel = scatter_dependence(dep);
    Relation R = pdg::map2relation(rel, params);
    rel->params.resize(0);
    rel->free();
    delete rel;
    return R;
}

static Matrix* Matrix2Matrix(pdg::Matrix *mat, int dim, int nparam)
{
    return mat->el.size() ? (Matrix*)*mat : Matrix_Alloc(0, 2+dim+nparam);
}

static pdg::UnionSet *scatter_node(pdg::node *node)
{
    int nparam = node->source->params.size();
    Polyhedron *domain = NULL;
    assert(node->source->dim == node->scattering->output);
    assert(node->source->params.size() == node->scattering->params.size());

    Matrix *scattering[node->scattering->constraints.size()];
    for (int j = 0; j < node->scattering->constraints.size(); ++j) {
	scattering[j] = *node->scattering->constraints[j];
    }

    for (int i = 0; i < node->source->constraints.size(); ++i) {
	Matrix *source = Matrix2Matrix(node->source->constraints[i], 
				       node->source->dim, nparam);
	int extra = source->NbColumns-2 - node->source->dim - nparam;
	for (int j = 0; j < node->scattering->constraints.size(); ++j) {
	    int scatdim = node->scattering->input + node->scattering->output;
	    assert(scattering[j]->NbColumns-2 == scatdim + nparam);
	    int cols = 2 + scatdim + extra + nparam;
	    Matrix *M = Matrix_Alloc(source->NbRows + scattering[j]->NbRows, cols);
	    for (int k = 0; k < scattering[j]->NbRows; ++k) {
		Vector_Copy(scattering[j]->p[k], M->p[k], 1+scatdim);
		Vector_Copy(scattering[j]->p[k]+1+scatdim, M->p[k]+1+scatdim+extra,
			    nparam+1);
	    }
	    int offset = scattering[j]->NbRows;
	    for (int k = 0; k < source->NbRows; ++k) {
		value_assign(M->p[offset+k][0], source->p[k][0]);
		Vector_Copy(source->p[k]+1, M->p[offset+k]+1+node->scattering->input,
			    source->NbColumns - 1);
	    }
	    Polyhedron *P = Constraints2Polyhedron(M, POL_NO_DUAL);
	    P->next = domain;
	    domain = P;
	    Matrix_Free(M);
	}
	Matrix_Free(source);
    }

    for (int j = 0; j < node->scattering->constraints.size(); ++j) {
	Matrix_Free(scattering[j]);
    }
    Polyhedron *S = Domain_projectout(domain, node->scattering->input,
				      node->scattering->output, nparam);
    Domain_Free(domain);
    pdg::UnionSet *scat_domain = new pdg::UnionSet(node->scattering->input,
						   S, node->source->params);
    Domain_Free(S);
    return scat_domain;
}

static void add_scat(cloog_program_builder& pb, int nparam, int dim, 
		     pdg::UnionSet *domain, int offset)
{
    int ddim = domain->constraints[0]->el[0].size() - 2 - nparam;
    Matrix *M = Matrix_Alloc(dim+1, 1+(dim+1)+ddim+nparam+1);
    for (int i = 0; i < dim; ++i) {
	value_set_si(M->p[i][1+i], 1);
	value_set_si(M->p[i][1+(dim+1)+i], -1);
    }
    value_set_si(M->p[dim][1+dim], 1);
    value_set_si(M->p[dim][1+(dim+1)+ddim+nparam], -offset);
    Polyhedron *scat = Constraints2Polyhedron(M, POL_NO_DUAL);
    Matrix_Free(M);
    
    pb.add_scattering(scat);
}

/* The ESPAM people want all "control" variables of the node domain
 * to appear in the port domains.
 * We therefore temporarily treat the control variables of the node domain
 * as iterators and then simplify the control variables of the port domain
 * in terms of these extra iterators.
 */
static pdg::UnionSet *normalize_domain(pdg::UnionSet *port_domain, 
				       pdg::UnionSet *node_domain,
				       const std::vector<Global_Var_Decl *>& params)
{
    int dim = node_domain->dim;
    int nparam = node_domain->params.size();
    unsigned nexist = domain_n_control(node_domain);
    if (!nexist)
	return port_domain;
    node_domain->dim += nexist;
    Relation domain = pdg::set2relation(node_domain, params);
    node_domain->dim -= nexist;
    Relation pd = pdg::set2relation(port_domain, params);
    pd = Extend_Set(pd, nexist);
    pd = Intersection(pd, domain);
    pd.simplify();
    port_domain = new pdg::OUnionSet(&pd, params);
    port_domain->dim -= nexist;
    port_domain->params = node_domain->params;
    return port_domain;
}

/* Write a linear combination of the (original) iterators and the parameters
 * as an expression in the new (scattered) iterators and the parameters.
 */
static void writeExpression(xmlTextWriterPtr writer, int i,
			    pdg::node *node, pdg::access *access)
{
    int nparam = node->source->params.size();
    Polyhedron *domain = NULL;
    assert(access->map->input == node->scattering->output);
    assert(access->map->params.size() == node->scattering->params.size());
    int dim = node->scattering->input;

    Matrix *scattering[node->scattering->constraints.size()];
    for (int j = 0; j < node->scattering->constraints.size(); ++j) {
	scattering[j] = *node->scattering->constraints[j];
    }

    assert(access->map->constraints.size() == 1);
    Matrix *map = (Matrix*)*access->map->constraints[0];

    assert(node->scattering->constraints.size() == 1);
    for (int j = 0; j < node->scattering->constraints.size(); ++j) {
	int scatdim = node->scattering->input + node->scattering->output;
	assert(scattering[j]->NbColumns-2 == scatdim + nparam);
	int cols = 2 + scatdim + 1 + nparam;
	Matrix *M = Matrix_Alloc(map->NbRows + scattering[j]->NbRows, cols);
	for (int k = 0; k < scattering[j]->NbRows; ++k) {
	    Vector_Copy(scattering[j]->p[k], M->p[k], 1+scatdim);
	    Vector_Copy(scattering[j]->p[k]+1+scatdim, M->p[k]+1+scatdim+1, nparam+1);
	}
	int offset = scattering[j]->NbRows;
	for (int k = 0; k < map->NbRows; ++k) {
	    value_assign(M->p[offset+k][0], map->p[k][0]);
	    Vector_Copy(map->p[k]+1, M->p[offset+k]+1+node->scattering->input,
			map->NbColumns - 1);
	}
	Polyhedron *P = Constraints2Polyhedron(M, POL_NO_DUAL);
	P->next = domain;
	domain = P;
	Matrix_Free(M);
    }
    Matrix_Free(map);
    Polyhedron *S = Domain_projectout(domain, node->scattering->input,
				      node->scattering->output, nparam);
    Domain_Free(domain);

    std::ostringstream strm;

    unsigned extra = S->Dimension - (dim + 1) - nparam;
    for (int row = 0; row < S->NbEq; ++row) {
	if (value_zero_p(S->Constraint[row][1+dim]))
	    continue;
	int sign = value_pos_p(S->Constraint[row][1+dim]) ? -1 : 1;
	value_absolute(S->Constraint[row][1+dim], S->Constraint[row][1+dim]);
	bool div = value_notone_p(S->Constraint[row][1+dim]);
	bool first = true;
	if (div)
	    strm << "(";
	for (int j = 0; j < dim; ++j) {
	    if (value_zero_p(S->Constraint[row][1+j]))
		continue;
	    if (!first && value_pos_p(S->Constraint[row][1+j]))
		strm << "+";
	    first = false;
	    strm << sign * VALUE_TO_INT(S->Constraint[row][1+j]) << "*" << "c" << j;
	}
	for (int j = 0; j < nparam; ++j) {
	    if (value_zero_p(S->Constraint[row][1+dim+1+extra+j]))
		continue;
	    if (!first && value_pos_p(S->Constraint[row][1+dim+1+extra+j]))
		strm << "+";
	    first = false;
	    strm << sign * VALUE_TO_INT(S->Constraint[row][1+dim+1+extra+j])
		 << "*" << node->source->params[j]->name->s;
	}
	if (value_notzero_p(S->Constraint[row][1+dim+1+extra+nparam])) {
	    if (!first && value_pos_p(S->Constraint[row][1+dim+1+extra+nparam]))
		strm << "+";
	    strm << sign * VALUE_TO_INT(S->Constraint[row][1+dim+1+extra+nparam]);
	}
	if (div)
	    strm << ")/" << VALUE_TO_INT(S->Constraint[row][1+dim]);
	break;
    }
    char *value = strdup(strm.str().c_str());

    Domain_Free(S);

    int rc;

    rc = xmlTextWriterStartElement(writer, BAD_CAST "expression");
    assert(rc >= 0);

    strm.str("");
    strm << "in_" << i;

    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name",
				     BAD_CAST strm.str().c_str());
    assert(rc >= 0);

    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "value", BAD_CAST value);
    assert(rc >= 0);

    rc = xmlTextWriterEndElement(writer);
    assert(rc >= 0);

    free(value);
}

static CloogProgram *writeADG(xmlTextWriterPtr writer, pdg::PDG *pdg)
{
    int rc;
    pdg::UnionSet **scattered_domains = NULL;

    int nparam = pdg->params.size();
    Polyhedron *ctx = pdg->context ? (Polyhedron *) *pdg->context : 
				     Universe_Polyhedron(nparam);

    cloog_program_builder pb(ctx);

    rc = xmlTextWriterStartElement(writer, BAD_CAST "adg");
    assert(rc >= 0);

    assert(pdg->name);
    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name", 
				     BAD_CAST pdg->name->s.c_str());
    assert(rc >= 0);

    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "levelUpNode", BAD_CAST "");
    assert(rc >= 0);

    for (int i = 0; i < pdg->params.size(); ++i) {
	rc = xmlTextWriterStartElement(writer, BAD_CAST "parameter");
	assert(rc >= 0);

	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name", 
					 BAD_CAST pdg->params[i]->name->s.c_str());
	assert(rc >= 0);

	assert(pdg->context->constraints.size() == 1);
	pdg::Matrix* M = pdg->context->constraints[0];
	for (int j = 0; j < M->el.size(); ++j) {
	    char buf[10];
	    if (!M->el[j][1+i])
		continue;
	    int k;
	    for (k = 0; k < pdg->params.size(); ++k)
		if (i != k && M->el[j][1+k])
		    break;
	    if (k < pdg->params.size())
		continue;
	    if (M->el[j][0] == 0) {
		int v = -M->el[j][1+pdg->params.size()]/M->el[j][1+i];
		snprintf(buf, sizeof(buf), "%d", v);
		rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "lb", 
						 BAD_CAST buf);
		assert(rc >= 0);
		rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "ub", 
						 BAD_CAST buf);
		assert(rc >= 0);
		continue;
	    }
	    if (M->el[j][1+i] == 1) {
		snprintf(buf, sizeof(buf), "%d", -M->el[j][1+pdg->params.size()]);
		rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "lb", 
						 BAD_CAST buf);
		assert(rc >= 0);
	    }
	    if (M->el[j][1+i] == -1) {
		snprintf(buf, sizeof(buf), "%d", M->el[j][1+pdg->params.size()]);
		rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "ub", 
						 BAD_CAST buf);
		assert(rc >= 0);
	    }
	}

	if (pdg->params[i]->value) {
	    char buf[10];
	    snprintf(buf, 10, "%d", pdg->params[i]->value->v);
	    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "value", BAD_CAST buf);
	    assert(rc >= 0);
	}

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);
    }

    scattered_domains = new pdg::UnionSet *[pdg->nodes.size()];
    for (int i = 0; i < pdg->nodes.size(); ++i)
	scattered_domains[i] = scatter_node(pdg->nodes[i]);

    vector<espam_edge*> split_edges;

    std::vector<Global_Var_Decl *> globals = pdg::extract_globals(pdg);
    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *dep = pdg->dependences[i];
	if (dep->type == pdg::dependence::uninitialized)
	    continue;
	if (dep->type == pdg::dependence::pn_part)
	    continue;
	if (dep->type == pdg::dependence::pn_wire)
	    continue;
	if (dep->type == pdg::dependence::pn_hold)
	    continue;
	vector<disjoint_edge> parts;

	if (dep->type == pdg::dependence::pn_union) {
	    for (int j = 0; j < pdg->dependences.size(); ++j) {
		pdg::dependence *part_dep = pdg->dependences[j];
		if (part_dep->type != pdg::dependence::pn_part)
		    continue;
		if (part_dep->container != dep)
		    continue;
		Relation remainder = scat_dep_relation(part_dep, globals);
		disjoin_dependences(parts, globals, remainder, part_dep->to_access,
				    false);
	    }
	} else {
	    Relation R = scat_dep_relation(dep, globals);
	    vector<pdg::access *> accesses;
	    accesses.push_back(dep->to_access);
	    parts.push_back(disjoint_edge(R, accesses, false));
	}

	for (int j = 0; j < pdg->dependences.size(); ++j) {
	    pdg::dependence *wire_dep = pdg->dependences[j];
	    if (wire_dep->type != pdg::dependence::pn_wire)
		continue;
	    if (dep->to != wire_dep->from)
		continue;
	    if (dep->array != wire_dep->array)
		continue;
	    if (dep->to_access != wire_dep->from_access)
		continue;
	    Relation dep_rel = scat_dep_relation(dep, globals);
	    Relation wire_rel = scat_dep_relation(wire_dep, globals);
	    Relation join = Join(dep_rel, wire_rel);
	    join.simplify();
	    if (join.is_satisfiable())
		disjoin_dependences(parts, globals, 
				    join, wire_dep->to_access, false);
	}

	for (int j = 0; j < pdg->dependences.size(); ++j) {
	    pdg::dependence *hold_dep = pdg->dependences[j];
	    if (hold_dep->type != pdg::dependence::pn_hold)
		continue;
	    if (dep->to != hold_dep->from)
		continue;
	    if (dep->array != hold_dep->array)
		continue;
	    if (dep->to_access != hold_dep->from_access)
		continue;
	    Relation dep_rel = scat_dep_relation(dep, globals);
	    Relation hold_rel = scat_dep_relation(hold_dep, globals);
	    Relation domain = Domain(hold_rel);
	    Relation sticky_part = Restrict_Range(dep_rel, domain);
	    if (sticky_part.is_satisfiable())
		disjoin_dependences(parts, globals, sticky_part, NULL, true);
	}

	for (int k = 0; k < parts.size(); ++k) {
	    if (!parts[k].domain.is_satisfiable())
		continue;
	    pdg::UnionMap *rel = new pdg::OUnionMap(&parts[k].domain, globals);
	    add_edge(pdg, split_edges, dep, rel, 
		     scattered_domains[dep->to->nr],
		     parts[k].accesses, parts[k].sticky, i, globals);
	}
    }

    div_vector dv;
    dv.index = 0;

    int max_access = 1;

    for (int i = 0; i < pdg->nodes.size(); ++i) {
	pdg::node *node = pdg->nodes[i];
	if (node->statement->accesses.size() > max_access)
	    max_access = node->statement->accesses.size();
    }

    for (int i = 0; i < pdg->nodes.size(); ++i) {
	char buf[10];
	char *name;
	pdg::node *node = pdg->nodes[i];

	rc = xmlTextWriterStartElement(writer, BAD_CAST "node");
	assert(rc >= 0);

	snprintf(buf, 10, "ND_%d", node->nr);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name", BAD_CAST buf);
	assert(rc >= 0);
	name = strdup(buf);

	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "levelUpNode", BAD_CAST "");
	assert(rc >= 0);

	vector<bool> is_write;
	vector<pdg::access*> accesses;
	for (int a = 0; a < node->statement->accesses.size(); ++a) {
	    int nr = node->statement->accesses[a]->nr;
	    if (nr >= is_write.size()) {
		is_write.resize(nr+1);
		accesses.resize(nr+1);
	    }
	    assert(nr < is_write.size());
	    if (node->statement->accesses[a]->type == pdg::access::write)
		is_write[nr] = true;
	    accesses[nr] = node->statement->accesses[a];
	}

	pdg::UnionSet *domain = scattered_domains[i];
	for (int i = 0, j = 0; i < split_edges.size(); ++i) {
	    if (split_edges[i]->to_node != node)
		continue;
	    char *port_name;
	    pdg::UnionSet *to_domain;
	    to_domain = split_edges[i]->to_domain;
	    port_name = writePort(writer, false, j, name, split_edges[i]->to_access,
				  split_edges[i],
				  to_domain, domain, pdg, dv);
	    split_edges[i]->to_port = port_name;
	    ++j;

	    pb.add_domain(*split_edges[i]->to_domain, node->statement->operation,
			  strdup(port_name));
	    add_scat(pb, nparam, pdg->dimension, split_edges[i]->to_domain,
		     (3*node->nr+0)*max_access+split_edges[i]->to_access[0]->nr);
	}

	for (int i = 0, j = 0; i < split_edges.size(); ++i) {
	    if (split_edges[i]->from_node != node)
		continue;
	    char *port_name;
	    pdg::UnionSet *from_domain;
	    from_domain = normalize_domain(split_edges[i]->from_domain, domain,
					   globals);
	    port_name = writePort(writer, true, j, name, split_edges[i]->from_access,
				  split_edges[i],
				  from_domain, domain, pdg, dv);
	    if (from_domain != split_edges[i]->from_domain)
		delete from_domain;
	    split_edges[i]->from_port = port_name;
	    ++j;

	    pb.add_domain(*split_edges[i]->from_domain, node->statement->operation,
			  strdup(port_name));
	    add_scat(pb, nparam, pdg->dimension, split_edges[i]->from_domain, 
		     (3*node->nr+2)*max_access+split_edges[i]->from_access[0]->nr);
	}

	for (int i = 0; i < accesses.size(); ++i)
	    if (!accesses[i]->array)
		writeExpression(writer, i, node, accesses[i]);

	rc = xmlTextWriterStartElement(writer, BAD_CAST "function");
	assert(rc >= 0);

	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name", 
		BAD_CAST (node->statement->top_function ?
			    node->statement->top_function->name->s.c_str() : ""));
	assert(rc >= 0);

	for (int i = 0; i < is_write.size(); ++i) {
	    snprintf(buf, 10, "%s_%d", is_write[i] ? "out" : "in", i);
	    rc = xmlTextWriterStartElement(writer, 
			BAD_CAST (is_write[i] ? "outargument" : "inargument"));
	    assert(rc >= 0);
	    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name", BAD_CAST buf);
	    assert(rc >= 0);
	    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "dataType", 
		 BAD_CAST(accesses[i]->array ? 
			    accesses[i]->array->element_type->s.c_str() : "int"));
	    assert(rc >= 0);
	    rc = xmlTextWriterEndElement(writer);
	    assert(rc >= 0);
	}

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);

	assert(pdg->dimension == node->scattering->input);
	writeLBS(writer, domain, domain, pdg, dv);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);

	dv.index += dv.divs.size();
	dv.divs.clear();

	pb.add_domain(*domain, node->statement->operation, name);
	add_scat(pb, nparam, pdg->dimension, domain, (3*node->nr+1)*max_access);

	domain->params.resize(0);
	domain->free();
	delete domain;
    }
    delete [] scattered_domains;

    for (int i = 0; i < split_edges.size(); ++i) {
	char buf[10];

	rc = xmlTextWriterStartElement(writer, BAD_CAST "edge");
	assert(rc >= 0);

	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name", 
					 BAD_CAST split_edges[i]->name);
	assert(rc >= 0);

	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "fromPort",
					 BAD_CAST split_edges[i]->from_port);
	assert(rc >= 0);

	snprintf(buf, 10, "ND_%d", split_edges[i]->from_node->nr);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "fromNode", BAD_CAST buf);
	assert(rc >= 0);

	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "toPort",
					 BAD_CAST split_edges[i]->to_port);
	assert(rc >= 0);

	snprintf(buf, 10, "ND_%d", split_edges[i]->to_node->nr);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "toNode", BAD_CAST buf);
	assert(rc >= 0);

	if (split_edges[i]->size)
	    snprintf(buf, 10, "%d", split_edges[i]->size->v);
	else
	    strcpy(buf, "-1");
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "size", BAD_CAST buf);
	assert(rc >= 0);

	rc = xmlTextWriterStartElement(writer, BAD_CAST "linearization");
	assert(rc >= 0);

	if (split_edges[i]->reordering)
	    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "type", 
					     BAD_CAST "GenericOutOfOrder");
	else if (split_edges[i]->multiplicity)
	    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "type", 
					     BAD_CAST "BroadcastInOrder");
	else if (split_edges[i]->sticky)
	    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "type", 
					     BAD_CAST "sticky_fifo");
	else if (split_edges[i]->shift_register)
	    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "type", 
					     BAD_CAST "shift_register");
	else
	    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "type", 
					     BAD_CAST "fifo");
	assert(rc >= 0);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);

	pdg::Matrix *stripped;
	stripped = strip_statement_dims_from_map(pdg, split_edges[i]->map);
	writeMatrix(writer, "mapping", stripped);
	delete stripped;

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);
    }

    rc = xmlTextWriterEndElement(writer);
    assert(rc >= 0);

    pdg->dimension++;
    pb.set_names(pdg->params, pdg->dimension);
    cloog_program_scatter(pb.p, pb.scatterings);

    return pb.p;
}

int main(int argc, char * argv[])
{
    char *input = NULL, *output = NULL;
    FILE *in = stdin, *out = stdout;
    int c, ind = 0;
    int rc;
    xmlTextWriterPtr writer;
    xmlDocPtr doc;

    LIBXML_TEST_VERSION

    while ((c = getopt_long(argc, argv, "i:o:", options, &ind)) != -1) {
	switch (c) {
	case 'i':
	    input = optarg;
	    break;
	case 'o':
	    output = optarg;
	    break;
	}
    }

    if (input && strcmp(input, "-")) {
	in = fopen(input, "r");
	assert(in);
	if (!output) {
	    int len = strlen(input);
	    if (len > 5 && !strcmp(input+len-5, ".yaml"))
		len -= 5;
	    output = new char[len+4+1];
	    strncpy(output, input, len);
	    strcpy(output+len, ".xml");
	}
    }

    PDG *pdg;
    pdg = yaml::Load<PDG>(in);

    writer = xmlNewTextWriterDoc(&doc, 0);
    assert(writer);

    rc = xmlTextWriterStartDocument(writer, NULL, NULL, NULL);
    assert(rc >= 0);

    rc = xmlTextWriterStartDTD(writer, BAD_CAST "sadg",
			    BAD_CAST "-//LIACS//DTD ESPAM 1//EN",
			    BAD_CAST "http://www.liacs.nl/~cserc/dtd/espam_1.dtd");
    assert(rc >= 0);
    rc = xmlTextWriterEndDTD(writer);
    assert(rc >= 0);

    rc = xmlTextWriterStartElement(writer, BAD_CAST "sadg");
    assert(rc >= 0);

    CloogProgram *p = writeADG(writer, pdg);

    writeAST(writer, p, pdg->dimension);

    rc = xmlTextWriterEndElement(writer);
    assert(rc >= 0);

    rc = xmlTextWriterEndDocument(writer);
    assert(rc >= 0);

    xmlFreeTextWriter(writer);

    if (output && strcmp(output, "-")) {
	out = fopen(output, "w");
	assert(out);
    }

    rc = xmlDocFormatDump(out, doc, 1);
    assert(rc >= 0);

    pdg->free();
    delete pdg;

    return 0;
}
