#include <iostream>
#include <ginac/ginac.h>
#include <barvinok/barvinok.h>
#include <barvinok/bernstein.h>
#include <bernstein/bernstein.h>
#include <bernstein/piecewise_lst.h>
extern "C" {
#include "combine.h"
#include "piputil.h"
}
#include "size_bernstein.h"
#include "evalue_util.h"

using namespace barvinok;
using namespace bernstein;
using namespace GiNaC;

namespace size {

using std::ostream;

ostream & operator<< (ostream & os, const exvector & v)
{
    os << "[";
    for (int i = 0; i < v.size(); ++i) {
	if (i)
	    os << ", ";
	os << v[i];
    }
    os << "]";
    return os;
}

/* Look for deepest loop dimension, i.e., a dimension that is not
 * fully determined by earlier dimensions.
 * To do this we perform a "reverse Gaussian elimination".
 * Starting from the innermost dimension, the first dimension 
 * that cannot be "eliminated" is the innermost loop dimension.
 *
 * dim: number of dimensions we want to consider
 * depth: current deepest loop dimension
 */
int update_depth(Polyhedron *P, int depth, int dim, Polyhedron *ctx)
{
    if (depth >= dim -1)
	return depth;
    int nparam = ctx->Dimension;
    int last = P->NbEq - 1;
    /* Matrix "view" of P for ExchangeRows */
    Matrix M;
    M.NbRows = P->NbConstraints;
    M.NbColumns = P->Dimension+2;
    M.p_Init = P->p_Init;
    M.p = P->Constraint;
    for (int d = P->Dimension-nparam; d > depth; --d) {
	int r;
	for (r = last; r >= 0; --r)
	    if (value_notzero_p(P->Constraint[r][1+d]))
		break;
	if (r < 0) {
	    if (d < dim) {
		depth = d;
		break;
	    }
	    continue;
	}
	if (r != last)
	    ExchangeRows(&M, r, last);
	for (--r; r >= 0; --r)
	    if (value_notzero_p(P->Constraint[r][1+d]))
		Combine(P->Constraint[r], P->Constraint[last],
			P->Constraint[r], 1+d, P->Dimension+1);
	--last;
    }
    return depth;
}

size::enumerator *selfloop_size_bernstein(pdg::PDG *pdg, Polyhedron *dep,
					  Polyhedron *reads,
					  int dim, Polyhedron *max,
					  Polyhedron *ctx,
					  size_options *options)
{
    size::enumerator *res = NULL;
    int nparam = ctx->Dimension;
    evalue *e = NULL;
    evalue mone;
    value_init(mone.d);
    evalue_set_si(&mone, -1, 1);

    piecewise_lst *pl_all = NULL;
    const char *param_names[nparam];
    Polyhedron *read;
    int depth = -1;
    for (Polyhedron *P = reads; P; P = P->next)
	depth = update_depth(P, depth, dim, ctx);
    for (Polyhedron *P = dep; P; P = P->next)
	depth = update_depth(P, depth, 2*dim, ctx);
    if (depth == -1)
	depth = 0;

    for (int i = 0; i < nparam; ++i)
	param_names[i] = pdg->params[i]->name->s.c_str();
    exvector params = constructParameterVector(param_names, nparam);

    for (read = reads; read; read = read->next) {
	int extra = read->Dimension - dim - nparam;
	Polyhedron *read_left = pip_lexminmax(read, depth, dim-depth+extra, nparam, 0);

	for (Polyhedron *RL = read_left; RL; RL = RL->next) {
	    evalue *EP = lexsmaller_domain(dep, RL, dim, ctx);
	    evalue *EC = lexsmaller_domain(reads, RL, dim, ctx);

	    emul(&mone, EC);
	    eadd(EC, EP);
	    free_evalue_refs(EC);
	    free(EC);
	    reduce_evalue(EP);

	    pl_all = evalue_bernstein_coefficients(pl_all, EP, ctx, params,
						   options->barvinok);
	    free_evalue_refs(EP);
	    free(EP);
	    if (!pl_all)
		break;
	}
	Domain_Free(read_left);
	if (!pl_all)
	    break;

	/*
	EP = evalue_max_domain(EP);
	if (!EP)
	    break;

	if (!evalue_is_constant(EP, dim)) {
	    free_evalue_refs(EP);
	    free(EP);
	    break;
	}

	assert(value_zero_p(EP->d));
	assert(EP->x.p->type == partition);
	assert(EP->x.p->size == 2);
	Polyhedron *D = EVALUE_DOMAIN(EP->x.p->arr[0]);
	Polyhedron *D_min = pip_lexminmax(D, 0, D->Dimension-nparam, nparam, 0);
	evalue_frac2floor(EP);
	evalue_remove_iterators(EP, D_min, dim, ctx);
	Polyhedron_Free(D_min);

	e = evalue_max(e, EP);

	if (!e)
	    break;
	*/
    }
    if (read && pl_all) {
	delete pl_all;
	pl_all = 0;
    }
    if (pl_all) {
	pl_all->maximize();
	pl_all->simplify_domains(ctx, options->barvinok->MaxRays);
	res = new size::enumerator(pl_all, pdg);
    }

    free_evalue_refs(&mone);

    return res;
}

}
