#include <iostream>
#include <set>
#include <limits.h>

#include "yaml.h"
#include "pdg.h"
#include "da.h"
extern "C" {
#include "piputil.h"
}
#include "translation.h"

using namespace std;

namespace pdg {

vector<int> operator+ (const vector<int> & v1, const vector<int> & v2)
{
    vector<int> v;
    v.resize(v1.size());
    for (int i = 0; i < v1.size(); ++i)
	v[i] = v1[i] + v2[i];
    return v;
}

void operator+= (vector<int> & v1, const vector<int> & v2)
{
    for (int i = 0; i < v1.size(); ++i)
	v1[i] += v2[i];
}

void operator-= (vector<int> & v1, const vector<int> & v2)
{
    for (int i = 0; i < v1.size(); ++i)
	v1[i] -= v2[i];
}

vector<int> operator- (const vector<int> & v)
{
    vector<int> vm;
    vm.resize(v.size());
    for (int i = 0; i < v.size(); ++i)
	vm[i] = -v[i];
    return vm;
}

template <typename T>
ostream & operator<< (ostream & os, const vector<T> & v)
{
    os << "[";
    for (int i = 0; i < v.size(); ++i) {
	if (i)
	    os << ", ";
	os << v[i];
    }
    os << "]";
    return os;
}

static void set_prefix(pdg::node *node, const vector<int>& prefix, int l)
{
    node->prefix.resize(0);
    for (int j = 0; j < l-1; ++j) {
	node->prefix.push_back(prefix[j]);
	node->prefix.push_back(-1);
    }
    node->prefix.push_back(prefix[l-1]);
}

/* Normalize prefix to be an alternation between statement
 * level dimensions and loop level dimensions.
 * Return the number of statement level dimension.
 */
int normalize_prefix(PDG *pdg)
{
    int i, l, maxl;
    vector<int> prefix;
    if (pdg->nodes.size() == 0)
	return 0;

    l = 1;
    for (int j = 0; j < pdg->nodes[0]->prefix.size(); ++j)
	if (pdg->nodes[0]->prefix[j] == -1)
	    ++l;
    prefix.resize(l);
    maxl = l;
    for (i = 1; i < pdg->nodes.size(); ++i) {
	int old_l = l;
	pdg::node *old_node = pdg->nodes[i-1];
	pdg::node *node = pdg->nodes[i];
	int change_l;
	l = 1;
	change_l = -1;
	for (int j = 0; j < node->prefix.size(); ++j) {
	    if (change_l == -1) {
		assert(old_node->prefix[j] <= node->prefix[j]);
		if (old_node->prefix[j] != node->prefix[j])
		    change_l = l;
	    }
	    if (node->prefix[j] == -1)
		++l;
	}
	if (l > maxl) {
	    maxl = l;
	    prefix.resize(l);
	}
	set_prefix(old_node, prefix, old_l);
	prefix[change_l-1]++;
	for (int j = change_l+1; j < prefix.size(); ++j)
	    prefix[j] = 0;
    }
    pdg::node *node = pdg->nodes[pdg->nodes.size()-1];
    set_prefix(node, prefix, l);
    return maxl;
}

void set_scattering_from_prefix(PDG *pdg)
{
    int statement_dims = normalize_prefix(pdg);
    int dim = 2 * statement_dims - 1;
    int nparam = pdg->params.size();

    assert(pdg->dimension == 0);
    assert(pdg->dimension == pdg->statement_dimensions.size());
    pdg->dimension = dim;
    for (int i = 0; i < dim; ++i)
	pdg->statement_dimensions.push_back(1-(i%2));

    for (int i = 0; i < pdg->nodes.size(); ++i) {
	pdg::node *node = pdg->nodes[i];
	pdg::UnionMap *s = new pdg::UnionMap;
	pdg::Matrix *M = new pdg::Matrix;
	s->input = dim;
	s->output = node->source->dim;
	s->params = pdg->params;
	int cols = 2+node->source->dim+nparam;
	if (node->source->constraints.size()) {
	    pdg::Matrix *SM = node->source->constraints[0];
	    if (SM->el.size() > 0)
		cols = SM->el[0].size();
	}
	M->el.insert(M->el.begin(), dim, std::vector<int>(dim+cols));
	for (int j = 0; j < statement_dims; ++j) {
	    M->el[2*j][1+2*j] = 1;
	    if (2*j < node->prefix.size())
		M->el[2*j][dim+cols-1] = -node->prefix[2*j];
	}
	for (int j = 0; j < statement_dims-1; ++j) {
	    M->el[2*j+1][1+2*j+1] = 1;
	    if (2*j+1 < node->prefix.size())
		M->el[2*j+1][1+dim+j] = -1;
	}
	s->constraints.push_back(M);
	node->scattering = s;
    }
}

/* Create a scattering function for the domains "source" based on "pos" and "scale".
 * The "extra" dimensions of source do not appear in the scattering function
 * since they have a unique value determined by the values of the "actual"
 * dimensions.
 */
static pdg::UnionMap *create_scattering(PDG *pdg, pdg::UnionSet *source,
				        vector<int>& pos, vector<int>& scale)
{
    int dim = pdg->dimension;
    int nparam = pdg->params.size();
    pdg::UnionMap *s = new pdg::UnionMap;
    pdg::Matrix *M = new pdg::Matrix;
    s->input = dim;
    s->output = source->dim;
    s->params = pdg->params;
    int cols = 2+source->dim+nparam;
    M->el.insert(M->el.begin(), dim, std::vector<int>(dim+cols));
    for (int i = 0; i < dim; ++i) {
	M->el[i][1+i] = 1;
	if (pos[i] >= 0)
	    M->el[i][1+dim+pos[i]] = -scale[i];
    }
    s->constraints.push_back(M);
    return s;
}

static pdg::UnionMap *derived_scattering(PDG *pdg, pdg::UnionSet *source,
			 pdg::UnionMap *scattering, pdg::UnionMap * rel, bool to2from)
{
    int dim = pdg->dimension;
    vector<int> pos(dim);
    vector<int> scale(dim);
    int this_dim = to2from ? rel->input : rel->output;
    int other_dim = to2from ? rel->output : rel->input; 
    // k: position in pos
    int k = 0, j = 0;
    for (j = 0; j < this_dim; ++j) {
	int l, m;

	/* l: source other dimension that corresponds to j */
	for (l = 0; l < other_dim; ++l) {
	    if (to2from && rel->constraints[0]->el[j][1+rel->input+l])
		break;
	    if (!to2from && rel->constraints[0]->el[l][1+rel->input+j])
		break;
	}
	if (l == other_dim)
	    break;
	/* m: scattering dimension that corresponds to l */
	for (m = 0; m < dim; ++m)
	    if (scattering->constraints[0]->el[m][1+dim+l])
		break;
	if (k > m)
	    break;
	if (m-j > dim - this_dim)
	    break;
	for ( ; k < m; ++k)
	    pos[k] = -1;
	if (!to2from)
	    scale[k] = scattering->constraints[0]->el[m][1+dim+l] *
		       rel->constraints[0]->el[l][1+rel->input+j];
	else {
	    if (scattering->constraints[0]->el[m][1+dim+l] % 
		rel->constraints[0]->el[j][1+rel->input+l] == 0)
		scale[k] = scattering->constraints[0]->el[m][1+dim+l] /
			    rel->constraints[0]->el[j][1+rel->input+l];
	    else scale[k] = 1;
	}
	pos[k++] = j;
    }
    for ( ; j < this_dim; ++j) {
	scale[k] = 1;
	pos[k++] = j;
    }
    for ( ; k < dim; ++k)
	pos[k] = -1;
    return create_scattering(pdg, source, pos, scale);
}

void common_dimension(PDG *pdg)
{
    int dim = pdg->dimension;
    for (int i = 0; i < pdg->nodes.size(); ++i) {
	pdg::node *node = pdg->nodes[i];
	if (node->source->dim > dim)
	    dim = node->source->dim;
    }
    assert(pdg->dimension == pdg->statement_dimensions.size());
    for (int i = pdg->dimension; i < dim; ++i)
	pdg->statement_dimensions.push_back(0);
    pdg->dimension = dim;
    vector<int> pos(dim);
    vector<int> scale(dim);
    for (int i = 0; i < dim; ++i) {
	scale[i] = 1;
	pos[i] = i;
    }

    int todo = pdg->nodes.size();
    for (int i = 0; i < pdg->nodes.size(); ++i) {
	pdg::node *node = pdg->nodes[i];
	if (node->source->dim != dim)
	    continue;
	node->scattering = create_scattering(pdg, node->source, pos, scale);
	--todo;
	break;
    }

    while (todo) {
	pdg::dependence *maxdep = NULL;
	vector<int> maxw(4);
	vector<int> w(4);
	maxw[0] = -1;
	maxw[1] = -1;
	maxw[2] = -1;
	maxw[3] = -1;
	w[0] = -1;
	w[1] = -1;
	w[2] = -1;
	w[3] = -1;
	for (int i = 0; i < pdg->dependences.size(); ++i) {
	    pdg::dependence *dep = pdg->dependences[i];
	    if (dep->type == pdg::dependence::uninitialized)
		continue;
	    pdg::UnionMap *rel = dep->relation;
	    if (dep->to->scattering && !dep->from->scattering) {
		w[0] = rel->output;
		w[1] = rel->input;
		w[2] = 1;
	    } else if (dep->from->scattering && !dep->to->scattering) {
		w[0] = rel->input;
		w[1] = rel->output;
		w[2] = 1;
		for (int j = 0; j < rel->input && j < rel->output; ++j)
		    w[2] *= rel->constraints[0]->el[j][1+rel->input+j];
		if (w[2] == 0)
		    w[2] = 1;
	    } else
		continue;
	    w[3] = dep->array->dims.size();
	    if (maxw < w) {
		maxw = w;
		maxdep = dep;
	    }
	}
	pdg::dependence *dep = maxdep;
	if (!dep) {
	    int maxi = -1;
	    int maxdim = -1;
	    for (int i = 0; i < pdg->nodes.size(); ++i) {
		pdg::node *node = pdg->nodes[i];
		if (node->scattering)
		    continue;
		if (node->source->dim > maxdim) {
		    maxdim = node->source->dim;
		    maxi = i;
		}
	    }
	    assert(maxi != -1);
	    pdg::node *node = pdg->nodes[maxi];
	    node->scattering = create_scattering(pdg, node->source, pos, scale);
	    fprintf(stderr, "WARNING: disconnected dependence graph\n");
	} else {
	    pdg::UnionMap *rel = dep->relation;
	    if (dep->to->scattering) {
		dep->from->scattering = derived_scattering(pdg, dep->from->source, 
				       dep->to->scattering, rel, true);
	    } else {
		dep->to->scattering = derived_scattering(pdg, dep->to->source, 
				       dep->from->scattering, rel, false);
	    }
	}
	--todo;
    }
}

struct dist {
    vector<int> v;
    int		fixed;

    dist() : fixed(0) {}
    dist(int v) : fixed(v) {}
    int sign(Polyhedron *ctx) const;
};

int dist::sign(Polyhedron *ctx) const
{
    return fixed;
}

ostream &operator<< (ostream &os, struct dist d)
{
    os << d.fixed;
    if (d.v.size() > 0)
	os << " (" << d.v << ")";
    return os;
}

vector<dist> operator+ (const vector<dist> & v1, const vector<int> & v2)
{
    vector<dist> v;
    v.resize(v1.size());
    for (int i = 0; i < v1.size(); ++i) {
	bool infinite = false;
	v[i].v.resize(v1[i].v.size());
	if (v1[i].v.size() > 0) {
	    int j;
	    for (j = 0; j < v1[i].v.size()-1; ++j)
		v[i].v[j] = v1[i].v[j];
	    v[i].v[j] = v1[i].v[j] + v2[i];
	}
	if (v1[i].fixed == INT_MAX || v1[i].fixed == -INT_MAX) {
	    infinite = true;
	    v[i].fixed = v1[i].fixed;
	}
	if (!infinite)
	    v[i].fixed = v1[i].fixed + v2[i];
    }
    return v;
}

void operator+= (dist& d1, const dist& d2)
{
    if (d1.v.size() > 0 && d2.v.size() == 0) {
	d1.v[d1.v.size()-1] += d2.fixed;
	if (d1.fixed != INT_MAX || d1.fixed != -INT_MAX)
	    d1.fixed += d2.fixed;
    } else if (d1.v.size() == 0 && d2.v.size() > 0) {
	d1.v.resize(d2.v.size());
	int j;
	for (j = 0; j < d2.v.size()-1; ++j)
	    d1.v[j] = d2.v[j];
	d1.v[j] = d1.fixed + d2.v[j];
	if (d2.fixed == INT_MAX || d2.fixed == -INT_MAX)
	    d1.fixed = d2.fixed;
	else
	    d1.fixed += d2.fixed;
    } else {
	assert(d1.v.size() == d2.v.size());
	for (int i = 0; i < d2.v.size(); ++i)
	    d1.v[i] += d2.v[i];
	if (d1.v.size() > 0) {
	    int i = 0;
	    for (i = 0; i < d1.v.size()-1; ++i)
		if (d1.v[i] != 0)
		    break;
	    if (i < d1.v.size()-1)
		assert(d1.fixed == d2.fixed);
	    else {
		d1.fixed = d1.v[d1.v.size()-1];
		d1.v.resize(0);
	    }
	} else
	    d1.fixed += d2.fixed;
    }
}

void operator+= (vector<dist>& v1, const vector<dist>& v2)
{
    for (int i = 0; i < v1.size(); ++i)
	v1[i] += v2[i];
}

vector<dist> operator+ (const vector<dist> & v1,
			    const vector<dist> & v2)
{
    vector<dist> v;
    v = v1;
    v += v2;
    return v;
}

bool operator< (const dist a, const dist b)
{
    assert(a.fixed != -INT_MAX && a.fixed != INT_MAX);
    assert(b.fixed != -INT_MAX && b.fixed != INT_MAX);
    return a.fixed < b.fixed;
}

bool operator<= (const dist a, const dist b)
{
    assert(a.v.size() == 0);
    assert(b.v.size() == 0);
    return a.fixed <= b.fixed;
}

bool operator== (const dist a, const dist b)
{
    if (a.v.size() != b.v.size())
	return false;
    for (int i = 0; i < a.v.size(); ++i)
	if (a.v[i] != b.v[i])
	    return false;
    return a.fixed == b.fixed;
}

dist operator- (dist d)
{
    dist dn;
    dn.v.resize(d.v.size());
    for (int i = 0; i < d.v.size(); ++i)
	dn.v[i] = -d.v[i];
    dn.fixed = - d.fixed;
    return dn;
}

void operator-= (dist& d1, const dist& d2)
{
    d1 += -d2;
}

dist operator- (const dist& d1, const dist& d2)
{
    dist dn = d1;
    dn -= d2;
    return dn;
}

void operator+= (dist& d1, int d2)
{
    if (d1.v.size() > 0)
	d1.v[d1.v.size()-1] += d2;
    d1.fixed += d2;
}

dist operator+ (dist d1, int d2)
{
    dist dn;
    dn.v.resize(d1.v.size());
    if (d1.v.size() > 0) {
	for (int i = 0; i < d1.v.size()-1; ++i)
	    dn.v[i] = d1.v[i];
	dn.v[dn.v.size()-1] = d1.v[d1.v.size()-1] + d2;
    }
    dn.fixed = d1.fixed + d2;
    return dn;
}

vector<dist> operator- (const vector<dist> & v)
{
    vector<dist> vm;
    vm.resize(v.size());
    for (int i = 0; i < v.size(); ++i)
	vm[i] = -v[i];
    return vm;
}

void operator-= (vector<dist>& v1, const vector<dist>& v2)
{
    for (int i = 0; i < v1.size(); ++i)
	v1[i] -= v2[i];
}

/*
 * Project on final dim dimensions
 */
static Polyhedron* Domain_Project(Polyhedron *P, int dim)
{
    int i;
    int remove = P->Dimension - dim;
    ::Matrix *T;
    Polyhedron *I;

    if (P->Dimension == dim)
	return Domain_Copy(P);

    T = Matrix_Alloc(dim+1, P->Dimension+1);
    for (i = 0; i < dim+1; ++i)
	value_set_si(T->p[i][i+remove], 1);
    I = DomainImage(P, T, P->NbConstraints);
    Matrix_Free(T);
    if (I->next) {
	Polyhedron *tmp = I;
	I = DomainConvex(I, I->NbRays);
	Domain_Free(tmp);
    }
    return I;
}

struct dvector {
    vector<dist> v;
    Polyhedron *ctx;

    dvector(int dim, int nparam);
    dvector(dvector *dv, dvector *offset);
    bool combine(dvector *dv, dvector *offset);
    void combine_context(Polyhedron *C);
    dvector *negation();

    ~dvector() {
	Polyhedron_Free(ctx);
    }

protected:
    dvector(Polyhedron *P, int nparam);
    dvector(Polyhedron *ctx);
};

/* A zero dvector over the whole parameter space */
dvector::dvector(int dim, int nparam)
{
    v.resize(dim);
    ctx = Universe_Polyhedron(nparam);
}

/* sum of the two dvectors */
dvector::dvector(dvector *dv, dvector *offset)
{
    v = dv->v + offset->v;
    ctx = DomainIntersection(dv->ctx, offset->ctx, 0);
}

/* replace by dv + offset if dv + offset is smaller
 * return true if dv + offset was smaller
 */
bool dvector::combine(dvector *dv, dvector *offset)
{
    Polyhedron *new_ctx = DomainIntersection(dv->ctx, offset->ctx, 0);
    combine_context(new_ctx);
    Polyhedron_Free(new_ctx);

    vector<dist> new_v = dv->v + offset->v;

    for (int i = 0; i < v.size(); ++i) {
	if (v[i] == new_v[i])
	    continue;
	dist diff = new_v[i] - v[i];
	int sign = diff.sign(ctx);
	if (sign == 0)
	    continue;
	if (sign < 0) {
	    v = new_v;
	    return true;
	}
	return false;
    }

    return false;
}

void dvector::combine_context(Polyhedron *C)
{
    Polyhedron *D = DomainUnion(ctx, C, 0);
    C = DomainConvex(D, D->NbRays);
    Domain_Free(D);
    Polyhedron_Free(ctx);
    ctx = C;
}

dvector::dvector(Polyhedron *P, int nparam)
{
    ctx = Domain_Project(P, nparam);
}

dvector::dvector(Polyhedron *C)
{
    ctx = Polyhedron_Copy(C);
}

dvector *dvector::negation()
{
    dvector *d = new dvector(ctx);
    d->v = -v;
    return d;
}

ostream & operator<< (ostream & os, struct dvector *dv)
{
    os << dv->v;
    return os;
}

struct wdvector : public dvector {
    int		w;
    bool	parametric;
    int		nfixed;

    wdvector(Polyhedron *P, int w, int dim, int nparam, bool statdims);
    wdvector(wdvector *dv, dvector * offset);
    bool combine(Polyhedron *P, int w, int dim, int nparam, bool statdims);
    void combine(wdvector *dv, dvector *offset);
    void add_statement_dims();

private:
    void assign(vector<dist>& v, Polyhedron *P, int dim, int nparam,
		bool statdims);
};

ostream & operator<< (ostream & os, struct wdvector *dv)
{
    os << dv->v;
    if (dv->parametric)
	os << "(" << dv->nfixed << ")";
    os << ", " << dv->w;
    return os;
}

template<typename T>
ostream & operator<< (ostream & os, map<pdg::node *, T>&m)
{
    typename map< pdg::node*, T>::iterator i;
    os << "{ ";
    for (i = m.begin(); i != m.end(); ++i)
	os << (*i).first->nr << ": " << (*i).second << "; ";
    os << " }" << endl;
    return os;
}

/* Check how many elements of the minimal vector d are nonnegative
 * or fixed (i.e., independent of parameters)
 * and returns this prefix of the minimal vector in min.
 * The return value is the number of elements in min.
 */
static int nonneg_param_lexmin(Polyhedron *D, int dim, vector<int>& min)
{
    vector<Polyhedron*> minpoly;
    for (Polyhedron *P = D; P; P = P->next)
	minpoly.push_back(P);
    for (int d = 0; d < dim; ++d) {
	vector<Polyhedron*> thispoly = minpoly;
	minpoly.clear();
	for (int p = 0; p < thispoly.size(); ++p) {
	    Polyhedron *P = thispoly[p];
	    /* Note: some of these "parameters" may actually be existential vars */
	    int nparam = P->Dimension - dim;
	    int v;
	    if (!F_ISSET(P, POL_VERTICES) &&
		    First_Non_Zero(P->Constraint[d]+1+dim, nparam) != -1) {
		POL_ENSURE_VERTICES(P);
		/* this shouldn't happen, but it does with piplib 1.3.6 */
		if (emptyQ(P))
		    continue;
	    }
	    if (First_Non_Zero(P->Constraint[d]+1+dim, nparam) == -1) {
		assert(value_zero_p(P->Constraint[d][0]));
		assert(value_one_p(P->Constraint[d][1+d]));
		v = -VALUE_TO_INT(P->Constraint[d][1+P->Dimension]);
	    } else {
		bool set = false;
		for (int i = 0; i < P->NbRays; ++i) {
		    /* A line */
		    if (value_zero_p(P->Ray[i][0]))
			return d;
		    /* A negative ray or vertex */
		    if (value_neg_p(P->Ray[i][1+d]))
			return d;
		    /* A positive ray */
		    if (value_zero_p(P->Ray[i][1+P->Dimension]))
			continue;
		    if (!set || value_cmp_si(P->Ray[i][1+d], v) < 0)
			v = VALUE_TO_INT(P->Ray[i][1+d]);
		    set = true;
		}
	    }
	    if (minpoly.size() == 0 || v < min[d]) {
		minpoly.clear();
		min[d] = v;
		minpoly.push_back(P);
	    } else if (v == min[d])
		minpoly.push_back(P);
	}
    }
    return dim;
}

void wdvector::assign(vector<dist>& v, Polyhedron *P, int dim, int nparam,
		     bool statdims)
{
    assert(P->NbEq >= dim);
    vector<int> min;
    min.resize(dim);
    int first_neg = nfixed = nonneg_param_lexmin(P, dim, min);
    if (!statdims && !P->next && nfixed == dim) {
	for (int i = 0; i < dim; ++i) {
	    assert(value_zero_p(P->Constraint[i][0]));
	    assert(value_one_p(P->Constraint[i][1+i]));
	    if (First_Non_Zero(P->Constraint[i]+1+dim, nparam) != -1) {
		nfixed = i;
		break;
	    }
	}
    }
    if (statdims || P->next || nfixed == dim) {
	parametric = false;
	if (statdims) {
	    v.resize(2*dim);
	    for (int i = 0; i < dim; ++i) {
		v[2*i].fixed = 0;
		v[2*i+1].fixed = min[i];
	    }
	    if (nfixed < dim)
		v[2*nfixed].fixed = -1;
	} else {
	    assert(nfixed == dim);
	    v.resize(dim);
	    for (int i = 0; i < dim; ++i) {
		v[i].fixed = min[i];
	    }
	}
    } else {
	assert(!statdims);
	parametric = true;
	for (int i = 0; i < dim; ++i) {
	    v.resize(dim);
	    assert(value_zero_p(P->Constraint[i][0]));
	    assert(value_one_p(P->Constraint[i][1+i]));
	    if (First_Non_Zero(P->Constraint[i]+1+dim, nparam) != -1) {
		v[i].fixed = min[i];
		v[i].v.resize(nparam+1);
		for (int j = 0; j < nparam; ++j)
		    v[i].v[j] = -VALUE_TO_INT(P->Constraint[i][1+dim+j]);
		v[i].v[nparam] = -VALUE_TO_INT(P->Constraint[i][1+P->Dimension]);
	    } else
		v[i].fixed = -VALUE_TO_INT(P->Constraint[i][1+P->Dimension]);
	}
	if (first_neg < dim)
	    v[first_neg].fixed = -INT_MAX;
    }
}

wdvector::wdvector(Polyhedron *P, int w, int dim, int nparam, bool statdims) : 
		    w(w), dvector(P, nparam)
{
    this->parametric = false;
    assign(v, P, dim, nparam, statdims);
}

wdvector::wdvector(wdvector *dv, dvector *offset) : dvector(dv, offset)
{
    parametric = dv->parametric;
    w = dv->w;
    nfixed = dv->nfixed;
}

bool wdvector::combine(Polyhedron *P, int new_w, int dim, int nparam, bool statdims)
{
    vector<dist> new_v;
    bool was_parametric = parametric;
    assign(new_v, P, dim, nparam, statdims);
    Polyhedron *new_ctx = Domain_Project(P, nparam);
    Polyhedron *D = DomainUnion(ctx, new_ctx, 0);
    Polyhedron_Free(new_ctx);
    new_ctx = DomainConvex(D, D->NbRays);
    Domain_Free(D);
    Polyhedron_Free(ctx);
    ctx = new_ctx;
    if (was_parametric || parametric) {
	if (!was_parametric || !parametric)
	    return false;
	for (int i = 0; i < dim; ++i) {
	    if (v[i].v.size() == 0 && new_v[i].v.size() == 0) {
		if (v[i].fixed == new_v[i].fixed)
		    continue;
		if (new_v[i].fixed < v[i].fixed)
		    v = new_v;
		break;
	    }
	    if (v[i].v.size() == 0 || new_v[i].v.size() == 0)
		return false;
	    for (int j = 0; j < nparam; ++j) {
		if (v[i].v[j] != new_v[i].v[j])
		    return false;
	    }
	    if (v[i].v[nparam] == new_v[i].v[nparam])
		continue;
	    if (new_v[i].v[nparam] < v[i].v[nparam])
		v = new_v;
	    break;
	}
	w += new_w;
	return true;
    }
    if (new_v < v)
	v = new_v;
    w += new_w;
    return true;
}

void wdvector::combine(wdvector *dv, dvector *offset)
{
    dvector::combine(dv, offset);
    w += dv->w;
}

void wdvector::add_statement_dims()
{
    vector<dist> new_v;
    new_v.resize(2*v.size());
    for (int i = 0; i < v.size(); ++i) {
	if (v[i].fixed == INT_MAX) {
	    new_v[2*i] = dist(1);
	    break;
	}
	if (v[i].fixed == -INT_MAX) {
	    new_v[2*i] = dist(-1);
	    break;
	}
	new_v[2*i] = dist(0);
	new_v[2*i+1] = dist(v[i].fixed);
    }
    if (parametric) {
	parametric = false;
	for (int i = 0; i < 2*v.size(); ++i)
	    new_v[i].v.resize(0);
    }
    v = new_v;
}

typedef map<pdg::node *, map<pdg::node *, wdvector *> >::iterator d_graph_iterator;

struct d_graph {
    int nparam;
    bool statdims;
    map<pdg::node *, map<pdg::node *, wdvector *> > data;

    d_graph(int nparam) {
	this->nparam = nparam;
	statdims = false;
    }

    void update(pdg::node *from, pdg::node *to, Polyhedron *ddv, int weight, int dim);
    void add_statement_dims();
};

ostream &operator<< (ostream &os, d_graph& graph)
{
    os << "nparam: " << graph.nparam << ", statdims: " << graph.statdims << endl;
    os << graph.data;
    return os;
}

void d_graph::add_statement_dims()
{
    d_graph_iterator i;
    map<pdg::node *, wdvector *>::iterator j;

    for (i = data.begin(); i != data.end(); ++i)
	for (j = (*i).second.begin(); j != (*i).second.end(); ++j)
	    j->second->add_statement_dims();
}

bool parametric_ddv(Polyhedron *ddv, int dim)
{
    vector<int> min;
    min.resize(dim);
    return nonneg_param_lexmin(ddv, dim, min) < dim;
}

void d_graph::update(pdg::node *from, pdg::node *to, Polyhedron *ddv,
		     int weight, int dim)
{
    if (data[from].find(to) == data[from].end())
	data[from][to] = new wdvector(ddv, weight, dim, nparam, statdims);
    else {
	if (!data[from][to]->combine(ddv, weight, dim, nparam, statdims)) {
	    assert(!statdims);
	    statdims = true;
	    add_statement_dims();
	    bool ok = data[from][to]->combine(ddv, weight, dim, nparam, statdims);
	    assert(ok);
	}
    }
}

struct combiner {
    d_graph& graph;
    /* dependence distance offsets used during moving */
    map<pdg::node *, dvector * > move_offsets;
    /* the relative offsets */
    map<pdg::node *, map<pdg::node *, dvector * > > delays;
    /* the original distance vectors */
    map<pdg::node *, map<pdg::node *, vector<dist> > > distances;
    /* the final offsets */
    map<pdg::node *, vector<dist> > offsets;
    int	    n_nodes;
    int	    dim;
    Polyhedron *ctx;
    int	    nparam;
    bool    move;

    combiner(d_graph& distances, int n_nodes, int dim, Polyhedron *ctx,
	     bool move) : 
	graph(distances), n_nodes(n_nodes), dim(dim), ctx(ctx),
	nparam(ctx->Dimension), move(move) {}
    ~combiner() {
	map<pdg::node *, map<pdg::node *, dvector * > >::iterator j;
	map<pdg::node *, dvector * >::iterator i;
	for (j = delays.begin(); j != delays.end(); ++j)
	    for (i = (*j).second.begin(); i != (*j).second.end(); ++i)
		delete (*i).second;
	for (i = move_offsets.begin(); i != move_offsets.end(); ++i)
	    delete (*i).second;
    }
    void combine();
    void compute_move_offsets();
    dvector *compute_relative_offset(pdg::node* from, pdg::node *to,
				     bool do_move);
    void set_relative_offset(pdg::node* from, pdg::node *to);
    void update_graph(pdg::node* from, pdg::node *to);
    void compute_absolute_offsets();
    void handle_zero_dependences();
};

dvector *combiner::compute_relative_offset(pdg::node* from, pdg::node *to,
					   bool do_move)
{
    dvector *delay;
    map<pdg::node *, dvector * > d;
    d[from] = new dvector(dim, nparam);

    bool progress = true;

    map<pdg::node *, dvector * >::iterator i;

    for (int k = 0; k < n_nodes && progress; ++k) {
	map<pdg::node *, wdvector *>::iterator j;
	progress = false;
	for (i = d.begin(); i != d.end(); ++i) {
	    pdg::node *a = (*i).first;
	    for (j = graph.data[a].begin(); j != graph.data[a].end(); ++j) {
		pdg::node *b = (*j).first;
		if (do_move && move_offsets.find(a) != move_offsets.end())
		    (*j).second->v -= move_offsets[a]->v;
		if (d.find(b) == d.end()) {
		    progress = true;
		    d[b] = new dvector((*j).second, d[a]);
		} else
		    progress = d[b]->combine((*j).second, d[a]) || progress;
		if (do_move && move_offsets.find(a) != move_offsets.end())
		    (*j).second->v += move_offsets[a]->v;
	    }
	}
    }

    /* d[from] + d[from] < d[from] iff d[from] < 0 */
    bool lexneg = d[from]->combine(d[from], d[from]);
    if (!do_move)
	assert(!lexneg);

    for (i = d.begin(); i != d.end(); ++i) {
	if ((*i).first == to)
	    continue;
	delete (*i).second;
    }

    delay = lexneg ? NULL : d[to]->negation();
    delete d[to];

    return delay;
}

void combiner::set_relative_offset(pdg::node* from, pdg::node *to)
{
    dvector *delay = NULL;
    if (move)
	delay = compute_relative_offset(from, to, true);
    if (!delay)
	delay = compute_relative_offset(from, to, false);
    delays[from][to] = delay;
}

/* Combine from and to at relative offset delays[from][to] */
void combiner::update_graph(pdg::node* from, pdg::node *to)
{
    dvector *offset = delays[from][to];
    dvector *neg_offset = offset->negation();
    map<pdg::node *, wdvector *>::iterator j;
    for (j = graph.data[to].begin(); j != graph.data[to].end(); ++j) {
	if ((*j).first != from) {
	    if (graph.data[from].find((*j).first) == graph.data[from].end())
		graph.data[from][(*j).first] = new wdvector((*j).second, neg_offset);
	    else
		graph.data[from][(*j).first]->combine((*j).second, neg_offset);
	}
	delete (*j).second;
    }
    delete neg_offset;
    graph.data.erase(to);
    d_graph_iterator i;
    for (i = graph.data.begin(); i != graph.data.end(); ++i) {
	if ((*i).second.find(to) == (*i).second.end())
	    continue;
	if ((*i).first != from) {
	    if ((*i).second.find(from) == (*i).second.end())
		(*i).second[from] = new wdvector((*i).second[to], offset);
	    else
		(*i).second[from]->combine((*i).second[to], offset);
	}
	delete (*i).second[to];
	(*i).second.erase(to);
    }
    --n_nodes;
}

void combiner::compute_absolute_offsets()
{
    offsets[(*graph.data.begin()).first].resize(dim);
    vector<pdg::node *> todo;
    todo.push_back((*graph.data.begin()).first);
    vector<dist> min(dim);
    while (!todo.empty()) {
	pdg::node *n = todo.back();
	todo.pop_back();
	if (delays.find(n) == delays.end())
	    continue;
	map<pdg::node *, dvector * >::iterator i;
	for (i = delays[n].begin(); i != delays[n].end(); ++i) {
	    offsets[(*i).first] = offsets[n] + (*i).second->v;
	    todo.push_back((*i).first);
	    for (int j = 0; j < dim; ++j) {
		if (offsets[(*i).first][j].fixed == INT_MAX)
		    continue;
		dist diff = offsets[(*i).first][j] - min[j];
		int sign = diff.sign(ctx);
		if (sign < 0)
		    min[j] = offsets[(*i).first][j];
	    }
	}
    }
    map<pdg::node *, vector<dist> >::iterator i;
    for (i = offsets.begin(); i != offsets.end(); ++i) {
	(*i).second -= min;
    }
}

static int lex_sign(const vector<dist>& v, Polyhedron *ctx)
{
    int sign = 0;
    for (int i = 0; i < v.size(); ++i)
	if ((sign = v[i].sign(ctx)))
	    break;
    return sign;
}

void combiner::handle_zero_dependences()
{
    map<pdg::node *, set<pdg::node *> > succ;
    map<pdg::node *, int > prec;
    map<pdg::node *, map<pdg::node *, vector<dist> > >::iterator i;
    map<pdg::node *, vector<dist> >::iterator j;
    for (j = offsets.begin(); j != offsets.end(); ++j)
	prec[(*j).first] = 0;

    for (i = distances.begin(); i != distances.end(); ++i)
	for (j = (*i).second.begin(); j != (*i).second.end(); ++j) {
	    distances[(*i).first][(*j).first] += offsets[(*j).first];
	    distances[(*i).first][(*j).first] -= offsets[(*i).first];
	    int sign = lex_sign(distances[(*i).first][(*j).first], ctx);
	    assert(sign >= 0);
	    if (sign == 0) {
		succ[(*i).first].insert((*j).first);
		prec[(*j).first]++;
	    }
	}
    for (j = offsets.begin(); j != offsets.end(); ++j)
	(*j).second.resize(dim+1);
    while (!prec.empty()) {
	map<pdg::node *, int >::iterator i;
	for (i = prec.begin(); (*i).second > 0; ++i) 
	    ;
	set<pdg::node *>::iterator j;
	for (j = succ[(*i).first].begin(); j != succ[(*i).first].end(); ++j) {
	    prec[(*j)]--;
	    if (offsets[(*j)][dim] <= offsets[(*i).first][dim])
		offsets[(*j)][dim] = offsets[(*i).first][dim] + 1;
	}
	prec.erase(i);
    }
}

static int node_dim(int *first, int *second)
{
    assert (*first == *second);
    return 2*(*first)-1;
}

/* Builds a relation that connects iterations of D to the next iteration */
static Polyhedron *next_iteration_relation(Polyhedron *D, int dim, Polyhedron *ctx)
{
    Polyhedron *res = NULL;
    int i, n = 0;
    for (Polyhedron *P = D; P; P = P->next)
	++n;

    pip_dependence_info *dep_info;
    dep_info = (pip_dependence_info *) malloc(sizeof(pip_dependence_info) + 
					      n * sizeof(pip_rel));
    dep_info->ctx = ctx;
    dep_info->precedes_level = (pip_access_precedes_level) node_dim;
    dep_info->n_write = n;
    i = 0;
    Polyhedron *next;
    for (Polyhedron *P = D; P; P = next, ++i) {
	next = P->next;
	P->next = NULL;
	dep_info->write[i].n_input = dim;
	dep_info->write[i].n_output = 0;
	dep_info->write[i].rel = P;
	dep_info->write[i].data = &dim;
    }
    dep_info->read.n_input = dim;
    dep_info->read.n_output = 0;
    dep_info->read.data = &dim;
    for (int i = 0; i < n; ++i) {
	dep_info->read.rel = dep_info->write[i].rel;
	pip_rel* pr = pip_lastwrite(dep_info);
	for (int j = 0; j < n; ++j) {
	    if (!pr[j].rel)
		continue;
	    res = Domain_concat(pr[j].rel, res);
	}
	Domain_Free(pr[n].rel);
	free(pr);
    }
    free(dep_info);
    return res;
}


void combiner::compute_move_offsets()
{
    d_graph_iterator i;
    map<pdg::node *, wdvector *>::iterator j;

    d_graph selfdistances(graph.nparam);

    for (i = graph.data.begin(); i != graph.data.end(); ++i) {
	for (j = (*i).second.begin(); j != (*i).second.end(); ++j) {
	    pdg::node *node = (*i).first;
	    pdg::UnionSet *source = node->source;
	    Polyhedron *D = *source;
	    Polyhedron *next = next_iteration_relation(D, source->dim, ctx);
	    pdg::UnionMap *scat = (*i).first->scattering;
	    Polyhedron *S = *scat;
	    assert(scat->output == source->dim);
	    for (Polyhedron *P = next; P; P = P->next) {
		Polyhedron *ddv = pip_min_ddv(P, S, S, source->dim, source->dim,
					      scat->input, ctx);
		selfdistances.update(node, node, ddv, 0, dim);
		Polyhedron_Free(ddv);
	    }
	    if (selfdistances.data[node][node])
		move_offsets[node] = selfdistances.data[node][node];
	    Polyhedron_Free(S);
	    Domain_Free(D);
	    Domain_Free(next);
	    break;
	}
    }
}

void combiner::combine()
{
    d_graph_iterator i;
    map<pdg::node *, wdvector *>::iterator j;

    if (move)
	compute_move_offsets();

    for (i = graph.data.begin(); i != graph.data.end(); ++i)
	for (j = (*i).second.begin(); j != (*i).second.end(); ++j)
	    distances[(*i).first][(*j).first] = (*j).second->v;

    while (n_nodes > 1) {
	int max = -1;
	pdg::node* max_from = NULL;
	pdg::node* max_to = NULL;
	d_graph_iterator i;
	map<pdg::node *, wdvector *>::iterator j;
	for (i = graph.data.begin(); i != graph.data.end(); ++i)
	    for (j = (*i).second.begin(); j != (*i).second.end(); ++j)
		if ((*j).second->w > max) {
		    max = (*j).second->w;
		    max_from = (*i).first;
		    max_to = (*j).first;
		}
	assert(max_from);
	assert(max_to);
	set_relative_offset(max_from, max_to);
	update_graph(max_from, max_to);
    }

    compute_absolute_offsets();
    handle_zero_dependences();
}

typedef pair<pdg::node *, pdg::node *> nodepair;

bool translate(PDG *pdg, bool move)
{
    int dim = pdg->dimension;
    int nparam = pdg->params.size();

    d_graph distances(nparam);
    Polyhedron *ctx = pdg->context ? (Polyhedron *) *pdg->context : 
				     Universe_Polyhedron(nparam);
    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *dep = pdg->dependences[i];
	if (dep->type == pdg::dependence::uninitialized)
	    continue;
	if (dep->from == dep->to)
	    continue;
	int weight = dep->type == pdg::dependence::anti ? 0 : 1;
	pdg::UnionMap *rel = dep->relation;
	pdg::node* from = dep->from;
	pdg::node* to = dep->to;
	Polyhedron *R = *rel;
	Polyhedron *sw = *(from->scattering);
	Polyhedron *sr = *(to->scattering);
	for (Polyhedron *P = R; P; P = P->next) {
	    Polyhedron *ddv = pip_min_ddv(P, sw, sr,
				      from->source->dim, to->source->dim, dim, ctx);
	    if (!ddv) {
		fprintf(stderr, "WARNING: empty dependence\n");
		continue;
	    }
	    for (Polyhedron *Q = ddv; Q; Q = Q->next) {
		Polyhedron *next = Q->next;
		Q->next = NULL;
		distances.update(from, to, Q, weight, dim);
		Q->next = next;
	    }
	    Domain_Free(ddv);
	}
	Domain_Free(R);
	Polyhedron_Free(sw);
	Polyhedron_Free(sr);
    }

    if (distances.statdims) {
	pdg->statement_dimensions.resize(2*dim);
	for (int i = dim-1; i >= 0; --i) {
	    pdg->statement_dimensions[2*i+1] = pdg->statement_dimensions[i];
	    pdg->statement_dimensions[2*i] = 1;
	}
	pdg->dimension = dim += dim;
    }

    set<pdg::node *> nodes_done;
    /* split the dependence graph into connected subgraphs
     * and compute offsets in each subgraph
     */
    while (nodes_done.size() < pdg->nodes.size()) {
	set<pdg::node *> current_nodes;
	/* a connected subgraph */
	d_graph current_distances(nparam);
	current_distances.statdims = distances.statdims;
	vector<pdg::node *> todo;

	for (int i = 0; i < pdg->nodes.size(); ++i) {
	    pdg::node *node = pdg->nodes[i];
	    if (nodes_done.find(node) != nodes_done.end())
		continue;
	    todo.push_back(node);
	    break;
	}
	while (!todo.empty()) {
	    pdg::node *node = todo.back();
	    todo.pop_back();
	    if (current_nodes.find(node) != current_nodes.end())
		continue;

	    current_nodes.insert(node);

	    d_graph_iterator i;
	    map<pdg::node *, wdvector * >::iterator j;
	    for (j = distances.data[node].begin();
		 j != distances.data[node].end(); ++j) {
		pdg::node *node2 = (*j).first;
		if (current_nodes.find(node2) != current_nodes.end())
		    continue;
		todo.push_back(node2);
	    }
	    for (i = distances.data.begin(); i != distances.data.end(); ++i) {
		pdg::node *node2 = (*i).first;
		if (current_nodes.find(node2) != current_nodes.end())
		    continue;
		if ((*i).second.find(node) == (*i).second.end())
		    continue;
		todo.push_back(node2);
	    }
	}

	set<pdg::node *>::iterator i;
	for (i = current_nodes.begin(); i != current_nodes.end(); ++i) {
	    current_distances.data[(*i)] = distances.data[(*i)];
	}

	combiner c(current_distances, current_nodes.size(), dim, ctx, move);
	c.combine();

	for (i = current_nodes.begin(); i != current_nodes.end(); ++i) {
	    pdg::node *node = *i;
	    nodes_done.insert(node);
	    pdg::UnionMap *s = node->scattering;
	    if (current_distances.statdims)
		s->input *= 2;
	    s->input++;
	    for (int j = 0; j < s->constraints.size(); ++j) {
		pdg::Matrix *M = s->constraints[j];
		vector< vector<int> >::iterator it;

		if (current_distances.statdims) {
		    for (int k = 0; k < M->el.size(); ++k)
			for (int d = dim/2-1; d >= 0; --d)
			    M->el[k].insert(M->el[k].begin() + 1 + d, 0);

		    int cols = M->el[0].size();
		    for (int d = dim/2-1; d >= 0; --d) {
			vector<int> row;
			row.resize(cols);
			row[1 + 2*d] = 1;
			M->el.insert(M->el.begin() + d, row);
		    }
		}

		/* column for innermost statement level dimension */
		int cols = M->el[0].size()+1;
		for (int k = 0; k < M->el.size(); ++k)
		    M->el[k].insert(M->el[k].begin() + 1 + dim, 0);

		/* constraint for innermost statement level dimension */
		vector<int> row;
		row.resize(cols);
		row[1 + dim] = 1;
		M->el.insert(M->el.begin() + dim, row);

		/* set offsets */
		for (int k = 0; k <= dim; ++k) {
		    if (c.offsets[node][k].v.size() == 0)
			M->el[k][cols-1] = -c.offsets[node][k].fixed;
		    else {
			int l;
			for (l = 0; l < nparam; ++l)
			    M->el[k][1+s->input+s->output+l] =
				-c.offsets[node][k].v[l];
			M->el[k][cols-1] = -c.offsets[node][k].v[l];
		    }
		}
	    }
	}
    }

    Polyhedron_Free(ctx);

    pdg->statement_dimensions.push_back(1);
    pdg->dimension++;

    return true;
}

}
