#include <sstream>
#include "xml_AST.h"

using namespace std;

static ostream& operator<< (ostream& os, clast_term *t);
static ostream& operator<< (ostream& os, clast_reduction *r);
static ostream& operator<< (ostream& os, clast_binary *b);

static ostream& operator<< (ostream& os, clast_expr *e)
{
    if (!e)
	return os;
    switch (e->type) {
    case clast_expr::expr_term:
	os << ((clast_term*) e);
	break;
    case clast_expr::expr_red:
	os << ((clast_reduction*) e);
	break;
    case clast_expr::expr_bin:
	os << ((clast_binary*) e);
	break;
    default:
	assert(0);
    }
    return os;
}

static ostream& operator<< (ostream& os, clast_term *t)
{
    os << VALUE_TO_INT(t->val);
    if (t->var)
	os << "*" << t->var;
    return os;
}

static ostream& operator<< (ostream& os, clast_reduction *r)
{
    if (r->n == 1) {
	os << (r->elts[0]);
	return os;
    }

    assert(r->type == clast_red_sum);
    os << (r->elts[0]);
    for (int i = 1; i < r->n; ++i) {
	assert(r->elts[i]->type == clast_expr::expr_term);
	clast_term *t = ((clast_term*) r->elts[i]);
	if (value_posz_p(t->val))
	    os << "+";
	os << (r->elts[i]);
    }
    return os;
}

static ostream& operator<< (ostream& os, clast_binary *b)
{
    switch (b->type) {
	case clast_bin_mod:
	    os << "mod(" << b->LHS << "," << VALUE_TO_INT(b->RHS) << ")";
	    return os;
	case clast_bin_div:
	    os <<  "(" <<  b->LHS << ")/(" << VALUE_TO_INT(b->RHS) << ")";
	    return os;
	case clast_bin_fdiv:
	    os << "div(" << b->LHS << "," << VALUE_TO_INT(b->RHS) << ")";
	    return os;
	case clast_bin_cdiv:
	    os << "-div(-(" << b->LHS << ")," << VALUE_TO_INT(b->RHS) << ")";
	    return os;
	default:
	    assert(0);
    }
}

static string expr2string(clast_expr *e)
{
    std::ostringstream strm;

    strm << e;

    return strm.str();
}

static void writeStmt(xmlTextWriterPtr writer, clast_stmt *s);

static void writeStmt(xmlTextWriterPtr writer, clast_for *f)
{
    int rc;
    char buf[10];

    rc = xmlTextWriterStartElement(writer, BAD_CAST "for");
    assert(rc >= 0);

    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "iterator",
				     BAD_CAST f->iterator);
    assert(rc >= 0);

    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "LB", 
				     BAD_CAST expr2string(f->LB).c_str());
    assert(rc >= 0);

    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "UB", 
				     BAD_CAST expr2string(f->UB).c_str());
    assert(rc >= 0);

    snprintf(buf, 10, "%d", VALUE_TO_INT(f->stride));
    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "stride", BAD_CAST buf);
    assert(rc >= 0);

    writeStmt(writer, f->body);

    rc = xmlTextWriterEndElement(writer);
    assert(rc >= 0);
}

static void writeStmt(xmlTextWriterPtr writer, clast_guard *g)
{
    int rc;
    char buf[10];

    assert(g->n == 1);

    rc = xmlTextWriterStartElement(writer, BAD_CAST "if");
    assert(rc >= 0);

    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "LHS", 
				     BAD_CAST expr2string(g->eq[0].LHS).c_str());
    assert(rc >= 0);

    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "RHS", 
				     BAD_CAST expr2string(g->eq[0].RHS).c_str());
    assert(rc >= 0);

    snprintf(buf, 10, "%d", g->eq[0].sign);
    rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "sign", BAD_CAST buf);
    assert(rc >= 0);

    writeStmt(writer, g->then);

    rc = xmlTextWriterEndElement(writer);
    assert(rc >= 0);
}

static void writeStmt(xmlTextWriterPtr writer, clast_user_stmt *u)
{
    int rc;
    char *name = (char *)u->statement->usr;

    if (!strchr(name, 'P')) {
	rc = xmlTextWriterStartElement(writer, BAD_CAST "stmt");
	assert(rc >= 0);

	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "node", BAD_CAST name);
	assert(rc >= 0);
    } else {
	rc = xmlTextWriterStartElement(writer, BAD_CAST "port");
	assert(rc >= 0);

	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name", BAD_CAST name);
	assert(rc >= 0);
    }

    rc = xmlTextWriterEndElement(writer);
    assert(rc >= 0);
}

static void writeStmt(xmlTextWriterPtr writer, clast_stmt *s)
{
    for ( ; s; s = s->next) {
	if (CLAST_STMT_IS_A(s, stmt_user)) {
	    writeStmt(writer, (clast_user_stmt *) s);
	} else if (CLAST_STMT_IS_A(s, stmt_for)) {
	    writeStmt(writer, (clast_for *) s);
	} else if (CLAST_STMT_IS_A(s, stmt_guard)) {
	    writeStmt(writer, (clast_guard *) s);
	}
    }
}

void writeAST(xmlTextWriterPtr writer, CloogProgram *p, int dim)
{
    int rc;

    rc = xmlTextWriterStartElement(writer, BAD_CAST "ast");
    assert(rc >= 0);

    CloogOptions *options = cloog_options_malloc() ;
    options->override = 1;
    options->f = -1;
    options->l = dim;
    options->esp = 1;
    options->sh = 1;

    p = cloog_program_generate(p, options);
    clast_stmt *stmt = cloog_util_create_clast(p, options);
    cloog_options_free(options);
    cloog_util_simple_guards(stmt);
    stmt = cloog_util_hoist_ifs(stmt);
    writeStmt(writer, stmt);
    cloog_clast_free(stmt);

    rc = xmlTextWriterEndElement(writer);
    assert(rc >= 0);
}
